from kubernetes import client, config
from omt.common import CmdTaskMixin


class KubernetesClient(CmdTaskMixin):
    def __init__(self, config_file=None):
        if not config_file:
            config.load_kube_config()
        else:
            config.load_kube_config(config_file)

        self.config_file = config_file

        corev1api = client.CoreV1Api()
        appsv1api = client.AppsV1Api()
        extensionsv1beta1api = client.ExtensionsV1beta1Api()
        self.client_instances = [corev1api, appsv1api, extensionsv1beta1api]

    def __getattr__(self, item: str):
        for one_instance in self.client_instances:
            if hasattr(one_instance, item):
                return getattr(one_instance, item)

    def get_namespace(self, resource_type: str, resource_name: str):
        list_namespaces = getattr(self, "list_%s_for_all_namespaces" % resource_type)
        all_namespaces = list_namespaces()
        for one in all_namespaces.items:
            if one.metadata.name == resource_name:
                return one.metadata.namespace

    ##################################
    ###### kubectl impl ##############
    ##################################

    def apply(self, file):
        # todo: since apply not supported, need to impl

        # option1: using kubectl instead
        # option2: upgrade k8s version to support server-side apply
        # option3: implemented in client side myself
        config = ' --kubeconfig %s ' % self.config_file if self.config_file else ''

        cmd = "kubectl %(config)s apply -f %(file)s" % locals()
        result = self.run_cmd(cmd)

    def edit(self, resource_type: str, resource_name: str, namespace: str):
        config = ' --kubeconfig %s ' % self.config_file if self.config_file else ''
        self.run_cmd("kubectl %(config)s edit %(resource_type)s %(resource_name)s --namespace %(namespace)s" % locals())

    def portforward(self):
        pass

    def download(self, resource_name, namespace, local_dir, remote_dir, container=None):
        config = ' --kubeconfig %s ' % self.config_file if self.config_file else ''
        container_options = '' if not container else '-c ' + container
        cmd = "kubectl %(config)s cp %(container_options)s %(namespace)s/%(resource_name)s:%(remote_dir)s %(local_dir)s" % locals()
        self.run_cmd(cmd)

    def upload(self, resource_name, namespace, local_dir, remote_dir, container=None):
        config = ' --kubeconfig %s ' % self.config_file if self.config_file else ''
        container_options = '' if not container else '-c ' + container
        cmd = "kubectl %(config)s cp %(container_options)s %(local_dir)s %(namespace)s/%(resource_name)s:%(remote_dir)s" % locals()
        self.run_cmd(cmd)

    def exec(self, resource_type, resource_name, namespace, command, container=None, stdin=True):
        config = ' --kubeconfig %s ' % self.config_file if self.config_file else ''
        interactive_option = '-it' if stdin else ''
        container_options = '' if not container else '-c ' + container
        cmd = 'kubectl %(config)s exec %(interactive_option)s %(resource_type)s/%(resource_name)s %(container_options)s --namespace %(namespace)s -- %(command)s' % locals()
        self.run_cmd(cmd)

    def get(self, resource_type, resource_name='', namespace='all', output='yaml'):
        resource_name = '' if not resource_name else resource_name
        config = ' --kubeconfig %s ' % self.config_file if self.config_file else ''
        namespace_options = ''
        if namespace == 'all':
            namespace_options = '-A'
        else:
            namespace_options = '--namespace %s' % namespace

        output_options = '' if not output else '-o ' + output

        cmd = 'kubectl %(config)s get %(resource_type)s %(resource_name)s %(namespace_options)s -o wide' % locals()
        result = self.run_cmd(cmd, capture_output=True, verbose=False)
        return result.stdout.decode('UTF-8')

    def describe(self, resource_type, resource_name=None, namespace='all'):
        config = ' --kubeconfig %s ' % self.config_file if self.config_file else ''
        namespace_options = ''
        if namespace == 'all':
            namespace_options = '-A'
        else:
            namespace_options = '--namespace %s' % namespace

        resource_name = '' if not resource_name else resource_name
        cmd = 'kubectl %(config)s describe %(resource_type)s %(resource_name)s %(namespace_options)s' % locals()
        result = self.run_cmd(cmd, capture_output=True, verbose=False)
        return result.stdout.decode("UTF-8")


if __name__ == '__main__':
    client = KubernetesClient("~/.omt/config/kube/nightly1/config")
    the_namespace = client.get_namespace("service", "smarta-smart-ticket-svc")
    print(the_namespace)
    #
    # print(client.list_config_map_for_all_namespaces(watch=False))
    # print(client.list_pod_for_all_namespaces(watch=False))
    # print(client.list_deployment_for_all_namespaces(watch=False))
    # print(client.list_service_for_all_namespaces(watch=False))
    # print(client.list_endpoints_for_all_namespaces(watch=False))



# -*- coding: utf-8 -*-
import os
from os.path import exists

from omt.config import settings

from omt.core.decorator import filecache

from omt.common import CmdTaskMixin
from omt.core import Resource
import pkg_resources

from omt.utils import JmxTermUtils


class Completion(Resource, CmdTaskMixin):


    def _description(self):
        return 'for resource completion'

    @filecache(duration=-1, file=os.path.join(settings.OMT_COMPLETION_CACHE_DIR, 'completion'))
    def _get_resource_completion(self):
        results = []
        resources = (pkg_resources.resource_listdir('omt', 'resources'))
        resources = list(filter(lambda x: x != '__init__.py' and x not in ['__pycache__'], resources))
        for resource_type in resources:
            mod = __import__(".".join(['omt', 'resources', resource_type, resource_type]),
                             fromlist=[resource_type.capitalize()])
            clazz = getattr(mod, resource_type.capitalize())
            results.append(resource_type + ":" + clazz({})._description())

        return "\n".join(results)

    def _run(self):
        if '--refresh' in self.context['all']:
            self._clean_completin_cache()
        print(self._get_resource_completion())





import pkg_resources
from urllib.parse import urlparse


class JmxTermUtils:
    @staticmethod
    def build_command(command):
        jmxterm = pkg_resources.resource_filename(__name__, '../lib/jmxterm-1.0.2-uber.jar')
        jmx_cmd = 'echo "%s"  | java -jar %s -n' % (command, jmxterm)
        return jmx_cmd


class UrlUtils:
    def __init__(self, url):
        self.url = url
        self.parsed_result = urlparse(url)

    def parse(self):
        return self.parsed_result

    def remove_identification(self):
        # 'http://guest:guest@localhost:15672/api'
        if self.parsed_result.username is not None:
            return self.parsed_result.geturl().replace(
                self.parsed_result.username + ":" + self.parsed_result.password + "@", "")
        else:
            return self.parsed_result.geturl()

    def get_username(self):
        return self.parsed_result.username

    def get_password(self):
        return self.parsed_result.password

    def get_hostname(self):
        return self.parsed_result.hostname

    def get_port(self):
        return self.parsed_result.port



import json
import os
from datetime import datetime

from omt.core.decorator import filecache

from omt.config import settings
from ruamel.yaml import YAML
from ruamel.yaml.compat import StringIO

from omt.common import CmdTaskMixin
from omt.core.resource import Resource
from omt.utils.file_utils import make_directory
from omt.utils.utils import get_obj_value, get_all_dict_Keys, set_obj_value, delete_obj_key


def dateconverter(o):
    if isinstance(o, datetime):
        return o.__str__()


class KubeResource(Resource, CmdTaskMixin):
    def __init__(self, context={}, type='web'):
        super().__init__(context, type)
        self.client = self.context['common']['client']

    def _get_kube_resource_type(self):
        return self._get_resource_name()

    def _read_namespaced_resource(self, name, namespace, **kwargs):
        read_func = getattr(self.client, 'read_namespaced_' + self._get_kube_resource_type())
        return read_func(name, namespace, **kwargs)

    def _list_resource_for_all_namespaces(self):
        list_func = getattr(self.client, 'list_%s_for_all_namespaces' % self._get_kube_resource_type())
        return list_func()

    @filecache(duration=60 * 5, file=Resource._get_cache_file_name)
    def _completion(self, short_mode=True):
        results = []
        results.append(super()._completion(False))

        if not self._have_resource_value():
            ret = self._list_resource_for_all_namespaces()
            results.extend(self._get_completion([one.metadata.name for one in ret.items], True))

        return "\n".join(results)

    def list(self):
        'display one or more resources'
        resource_name = self._get_one_resource_value()
        namespace = 'all' if not resource_name else self.client.get_namespace(self._get_kube_resource_type(),
                                                                              resource_name)

        # ret = self._list_resource_for_all_namespaces()
        # print(ret)
        result = self.client.get(self._get_kube_resource_type(), resource_name, namespace)
        print(result)

    def yaml(self):
        'get configuration in yaml format'
        resource = self._get_one_resource_value()
        namespace = self.client.get_namespace(self._get_kube_resource_type(), resource)
        result = self._read_namespaced_resource(resource, namespace)
        stream = StringIO()
        the_result = result.to_dict()
        yaml = YAML()
        yaml.dump(the_result, stream)
        print(stream.getvalue())

    def json(self):
        'get configuration by json format'
        resource = self._get_one_resource_value()
        namespace = self.client.get_namespace(self._get_kube_resource_type(), resource)
        result = self._read_namespaced_resource(resource, namespace)
        the_result = result.to_dict()
        print(json.dumps(the_result, default=dateconverter, indent=4))

    @staticmethod
    def _build_field_selector(selectors):
        return ','.join(['%s=%s' % (k, v) for (k, v) in selectors.items()])

    def namespace(self):
        'get resource namespace'
        resource = self._get_one_resource_value()
        namespace = self.client.get_namespace(self._get_kube_resource_type(), resource)
        print(namespace)

    def event(self):
        'show event on the resource'
        # https://kubernetes.docker.internal:6443/api/v1/namespaces/default/events?fieldSelector=
        # involvedObject.uid=4bb31f4d-99f1-4acc-a024-8e2484573733,
        # involvedObject.name=itom-xruntime-rabbitmq-6464654786-vnjxz,
        # involvedObject.namespace=default

        resource = self._get_one_resource_value()
        namespace = self.client.get_namespace(self._get_kube_resource_type(), resource)
        result = self._read_namespaced_resource(resource, namespace)

        uid = get_obj_value(result, 'metadata.uid')
        name = get_obj_value(result, 'metadata.name')

        the_selector = {
            "involvedObject.uid": uid,
            "involvedObject.namespace": namespace,
            "involvedObject.name": name,
        }

        print(self.client.list_namespaced_event(namespace, field_selector=self._build_field_selector(the_selector)))

    def _get_config_key_cache_file_name(self):
        main_path = [one for one in self.context['all'][1:] if not one.startswith('-')]
        cache_file = os.path.join(settings.OMT_COMPLETION_CACHE_DIR, *main_path)
        return cache_file

    @filecache(duration=60 * 5, file=_get_config_key_cache_file_name)
    def _get_config_key_completion(self):
        resource = self._get_one_resource_value()
        namespace = self.client.get_namespace(self._get_kube_resource_type(), resource)
        result = self._read_namespaced_resource(resource, namespace)
        prompts = []
        get_all_dict_Keys(result.to_dict(), prompts)
        return '\n'.join(self._get_completion(prompts))

    def get(self):
        'get resource by configuration key'
        if 'completion' in self._get_params():
            completion = self._get_config_key_completion()
            print(completion)
            return

        resource = self._get_one_resource_value()
        namespace = self.client.get_namespace(self._get_kube_resource_type(), resource)
        result = self._read_namespaced_resource(resource, namespace)
        params = self._get_action_params()

        the_params = " ".join(params)

        if not the_params.strip():
            print(result)
        else:
            print(get_obj_value(result, the_params))

    def set(self):
        'update restore by configuration key'
        if 'completion' in self._get_params():
            resource = self._get_one_resource_value()
            namespace = self.client.get_namespace(self._get_kube_resource_type(), resource)
            result = self._read_namespaced_resource(resource, namespace)
            prompts = []
            get_all_dict_Keys(result.to_dict(), prompts)
            self._print_completion(prompts)
            return

        resource = self._get_one_resource_value()
        namespace = self.client.get_namespace(self._get_kube_resource_type(), resource)
        result = self._read_namespaced_resource(resource, namespace)
        params = self._get_action_params()
        config_key = params[0]
        config_value = params[1]
        orig_value = get_obj_value(result, config_key)
        # convert type
        config_value = type(orig_value)(config_value)
        set_obj_value(result, config_key, config_value)

        # todo: use apply instead once apply provided
        new_result = self.client.replace_namespaced_deployment(resource, namespace, result)
        print(get_obj_value(new_result, config_key))

    def delete(self):
        'delete node by configuration key'
        # todo@rain: to support delete entired resource and completion cache
        if 'completion' in self._get_params():
            resource = self._get_one_resource_value()
            namespace = self.client.get_namespace(self._get_kube_resource_type(), resource)
            result = self._read_namespaced_resource(resource, namespace)
            prompts = []
            get_all_dict_Keys(result.to_dict(), prompts)
            self._print_completion(prompts)
            return

        resource = self._get_one_resource_value()
        namespace = self.client.get_namespace(self._get_kube_resource_type(), resource)
        result = self._read_namespaced_resource(resource, namespace)
        params = self._get_action_params()
        config_key = params[0]
        # convert type
        delete_obj_key(result, config_key)

        # todo: use apply instead once apply provided
        new_result = self.client.replace_namespaced_deployment(resource, namespace, result)

    def edit(self):
        'Edit a resource from the default editor.'
        resource = self._get_one_resource_value()
        namespace = self.client.get_namespace(self._get_kube_resource_type(), resource)

        self.client.edit(self._get_kube_resource_type(), resource, namespace)

    def save(self):
        'save configuration in file cache to be restored'
        resource_name = self._get_one_resource_value()
        namespace = self.client.get_namespace(self._get_kube_resource_type(), resource_name)
        kube_instance = self._get_one_resource_value("kube")
        if not kube_instance:
            kube_instance = 'local'
        cache_folder = os.path.join(settings.OMT_KUBE_CACHE_DIR, kube_instance, namespace,
                                    self._get_kube_resource_type())

        result = self._read_namespaced_resource(resource_name, namespace, _preload_content=False)
        stream = StringIO()
        the_result = json.loads(result.data.decode('UTF-8'))
        delete_obj_key(the_result, 'metadata.creationTimestamp')
        delete_obj_key(the_result, 'metadata.resourceVersion')
        yaml = YAML()
        yaml.dump(the_result, stream)
        content = stream.getvalue()

        make_directory(cache_folder)
        with open(os.path.join(cache_folder, resource_name + '.yaml'), 'w') as f:
            f.write(content)

    def restore(self):
        'restore configuration saved in file cache'
        resource_name = self._get_one_resource_value()
        namespace = self.client.get_namespace(self._get_kube_resource_type(), resource_name)
        kube_instance = self._get_one_resource_value("kube")
        if not kube_instance:
            kube_instance = 'local'
        cache_folder = os.path.join(settings.OMT_KUBE_CACHE_DIR, kube_instance, namespace,
                                    self._get_kube_resource_type())
        make_directory(cache_folder)

        config_file = os.path.join(cache_folder, resource_name + '.yaml')
        if os.path.exists(config_file):
            self.client.apply(config_file)
        else:
            raise Exception("no config file found")

    def exec(self):
        'Execute a command in a container'
        resource_name = self._get_one_resource_value()
        namespace = self.client.get_namespace(self._get_kube_resource_type(), resource_name)

        self.client.exec(self._get_kube_resource_type(), resource_name, namespace, " ".join(self._get_action_params()))

    def describe(self):
        'Show details of a specific resource or group of resources'
        resource_name = self._get_one_resource_value()
        namespace = 'all'
        if resource_name:
            namespace = self.client.get_namespace(self._get_kube_resource_type(), resource_name)

        print(self.client.describe(self._get_kube_resource_type(), resource_name, namespace))



import os


def make_directory(dir):
    if not os.path.exists(dir):
        os.makedirs(dir)


def inplace_replace(file_name, old_string, new_string):
    with open(file_name) as f:
        new_content = f.read().replace(old_string, new_string)

    with open(file_name, "w") as f:
        f.write(new_content)



import pkg_resources

from omt.common import CmdTaskMixin
from omt.core import Resource


class Config(Resource, CmdTaskMixin):
    def init(self):
        # 1. set up omt-completion.sh
        completion = pkg_resources.resource_filename(__name__, '../../../lib/jmxterm-1.0.2-uber.jar')



    def sync_version(self):
        pass



import functools
import os


def rsetattr(obj, attr, val):
    pre, _, post = attr.rpartition('.')
    return setattr(rgetattr(obj, pre) if pre else obj, post, val)


sentinel = object()


def rgetattr(obj, attr, default=sentinel):
    if default is sentinel:
        _getattr = getattr
    else:
        def _getattr(obj, name):
            return getattr(obj, name, default)
    return functools.reduce(_getattr, [obj] + attr.split('.'))


def omt_import(name):
    components = name.split('.')

    for comp in components:
        mod = __import__(components[0])
        mod = getattr(mod, comp)
    return mod


def make_directory(dir):
    if not os.path.exists(dir):
        os.makedirs(dir)


def get_obj_value2(obj, key):
    return functools.reduce(lambda x, y: getattr(x, y), key.split('.'), obj)


def delete_obj_key(obj, key):
    if not key:
        return None

    first_attr, others = extract_first_attr(key)

    if not others:
        # do set value

        if first_attr:
            if isinstance(obj, dict):
                # obj.pop(key, None)
                del obj[key]
            elif isinstance(obj, list):
                del obj[int(first_attr)]
            else:
                try:
                    delattr(obj, first_attr)
                except:
                    setattr(obj, first_attr, None)

            return
    else:
        # do get value

        if first_attr:
            if isinstance(obj, dict):
                first_value = obj.get(first_attr)
            elif isinstance(obj, list):
                first_value = obj[int(first_attr)]
            else:
                first_value = getattr(obj, first_attr)

            delete_obj_key(first_value, others)


def set_obj_value(obj, key, value):
    # e.g. get pod.data.ips[0]
    if not key:
        return None

    first_attr, others = extract_first_attr(key)

    if not others:
        # do set value

        if first_attr:
            if isinstance(obj, dict):
                obj[first_attr] = value
            elif isinstance(obj, list):
                obj[int(first_attr)] = value
            else:
                setattr(obj, first_attr, value)

            return
    else:
        # do get value

        if first_attr:
            if isinstance(obj, dict):
                first_value = obj.get(first_attr)
            elif isinstance(obj, list):
                first_value = obj[int(first_attr)]
            else:
                first_value = getattr(obj, first_attr)

            set_obj_value(first_value, others, value)


def get_obj_value(obj, key):
    # e.g. get pod.data.ips[0]
    if not key:
        return None

    first_attr, others = extract_first_attr(key)

    if first_attr:
        if isinstance(obj, dict):
            first_value = obj.get(first_attr)
        elif isinstance(obj, list):
            first_value = obj[int(first_attr)]
        else:
            first_value = getattr(obj, first_attr)

        if not others:
            return first_value

        else:
            return get_obj_value(first_value, others)


def extract_first_attr(key):
    key = key.strip('[].')
    for index in range(0, len(key)):
        if key[index] in '[].':
            return key[:index], key[index:].strip('[].')

    return key, None


def get_all_dict_Keys(obj, paths=[]):
    if isinstance(obj, dict):
        for k, v in obj.items():
            paths.append(k)
            subpaths = []
            get_all_dict_Keys(v, subpaths)
            paths.extend([k + one if one.startswith('[') else k + '.' + one for one in subpaths])
    elif isinstance(obj, list):
        paths.append('[0]')
        subpaths = []
        get_all_dict_Keys(obj[0], subpaths)
        paths.extend(['[0].' + one for one in subpaths])


if __name__ == '__main__':
    import json

    obj = json.loads('{"a":"a", "b": [{"e": "sdfsdf"}, "d"]}')
    set_obj_value(obj, 'b[0].e', 'test')
    delete_obj_key(obj, 'b[0].e')
    print(obj)
    print(get_obj_value(obj, 'b[0].e'))

    paths = []
    get_all_dict_Keys(obj, paths)
    print(paths)
    # first,other = (extract_first_attr('a[1].b'))
    # second,other2 = extract_first_attr(other)
    # print(extract_first_attr(other2))



import json
import sys

def maybe_utf8(s):
    if isinstance(s, int):
        # s can be also an int for ex messages count
        return str(s)
    if isinstance(s, float):
        # s can be also a float for message rate
        return str(s)
    if sys.version_info[0] == 3:
        # It will have an encoding, which Python will respect
        return s
    else:
        # It won't have an encoding, and Python will pick ASCII by default
        return s.encode('utf-8')


class Lister:
    def verbose(self, string):
        if self.options.verbose:
            print(string)

    def display(self, json_list):
        depth = sys.maxsize
        if len(self.columns) == 0:
            depth = int(self.options.depth)
        (columns, table) = self.list_to_table(json.loads(json_list), depth)
        if len(table) > 0:
            self.display_list(columns, table)
        else:
            self.verbose("No items")

    def list_to_table(self, items, max_depth):
        columns = {}
        column_ix = {}
        row = None
        table = []

        def add(prefix, depth, item, fun):
            for key in item:
                column = prefix == '' and key or (prefix + '.' + key)
                subitem = item[key]
                if type(subitem) == dict:
                    fun(column, json.dumps(subitem))
                    # if 'json' in self.obj_info and key in self.obj_info['json']:
                    #     fun(column, json.dumps(subitem))
                    # else:
                    #     if depth < max_depth:
                    #         add(column, depth + 1, subitem, fun)
                elif type(subitem) == list:
                    # The first branch has mirrors in queues in
                    # mind (which come out looking decent); the second
                    # one has applications in nodes (which look less
                    # so, but what would look good?).
                    if [x for x in subitem if type(x) != str] == []:
                        serialised = " ".join(subitem)
                    else:
                        serialised = json.dumps(subitem)
                    fun(column, serialised)
                else:
                    fun(column, subitem)

        def add_to_columns(col, val):
            columns[col] = True

        def add_to_row(col, val):
            if col in column_ix:
                if val is not None:
                    row[column_ix[col]] = maybe_utf8(val)
                else:
                    row[column_ix[col]] = None

        if len(self.columns) == 0:
            for item in items:
                add('', 1, item, add_to_columns)
            columns = list(columns.keys())
            columns.sort(key=column_sort_key)
        else:
            columns = self.columns

        for i in range(0, len(columns)):
            column_ix[columns[i]] = i
        for item in items:
            row = len(columns) * ['']
            add('', 1, item, add_to_row)
            table.append(row)

        return (columns, table)


class TSVList(Lister):
    def __init__(self, columns, obj_info, options):
        self.columns = columns
        self.obj_info = obj_info
        self.options = options

    def display_list(self, columns, table):
        head = "\t".join(columns)
        self.verbose(head)

        for row in table:
            line = "\t".join(row)
            print(line)


class LongList(Lister):
    def __init__(self, columns, obj_info, options):
        self.columns = columns
        self.obj_info = obj_info
        self.options = options

    def display_list(self, columns, table):
        sep = "\n" + "-" * 80 + "\n"
        max_width = 0
        for col in columns:
            max_width = max(max_width, len(col))
        fmt = "{0:>" + str(max_width) + "}: {1}"
        print(sep)
        for i in range(0, len(table)):
            for j in range(0, len(columns)):
                print(fmt.format(columns[j], table[i][j]))
            print(sep)


class TableList(Lister):
    def __init__(self, columns):
        self.columns = columns

    def display_list(self, columns, table):
        total = [columns]
        total.extend(table)
        self.ascii_table(total)

    def ascii_table(self, rows):
        col_widths = [0] * len(rows[0])
        for i in range(0, len(rows[0])):
            for j in range(0, len(rows)):
                col_widths[i] = max(col_widths[i], len(rows[j][i]))
        self.ascii_bar(col_widths)
        self.ascii_row(col_widths, rows[0], "^")
        self.ascii_bar(col_widths)
        for row in rows[1:]:
            self.ascii_row(col_widths, row, "<")
        self.ascii_bar(col_widths)

    def ascii_row(self, col_widths, row, align):
        txt = "|"
        for i in range(0, len(col_widths)):
            fmt = " {0:" + align + str(col_widths[i]) + "} "
            txt += fmt.format(row[i]) + "|"
        print(txt)

    def ascii_bar(self, col_widths):
        txt = "+"
        for w in col_widths:
            txt += ("-" * (w + 2)) + "+"
        print(txt)


class KeyValueList(Lister):
    def __init__(self, columns, obj_info, options):
        self.columns = columns
        self.obj_info = obj_info
        self.options = options

    def display_list(self, columns, table):
        for i in range(0, len(table)):
            row = []
            for j in range(0, len(columns)):
                row.append("{0}=\"{1}\"".format(columns[j], table[i][j]))
            print(" ".join(row))


# TODO handle spaces etc in completable names
class BashList(Lister):
    def __init__(self, columns, obj_info, options):
        self.columns = columns
        self.obj_info = obj_info
        self.options = options

    def display_list(self, columns, table):
        ix = None
        for i in range(0, len(columns)):
            if columns[i] == 'name':
                ix = i
        if ix is not None:
            res = []
            for row in table:
                res.append(row[ix])
            print(" ".join(res))


FORMATS = {
    # Special cased
    'raw_json': None,
    # Ditto
    'pretty_json': None,
    'tsv': TSVList,
    'long': LongList,
    'table': TableList,
    'kvp': KeyValueList,
    'bash': BashList
}

def format_list(json_list, columns=[], format='pretty_json'):
    formatter = None
    if format == "raw_json":
        print(json_list)
        return
    elif format == "pretty_json":
        #enc = json.JSONEncoder(False, False, True, True, True, indent=2)
        enc = json.JSONEncoder(indent=2)
        print(enc.encode(json.loads(json_list)))
        return
    else:
        formatter = FORMATS[format]
    formatter_instance = formatter(columns)
    formatter_instance.display(json_list)


#!/usr/bin/env python

#   The contents of this file are subject to the Mozilla Public License
#   Version 1.1 (the "License"); you may not use this file except in
#   compliance with the License. You may obtain a copy of the License at
#   https://www.mozilla.org/MPL/
#
#   Software distributed under the License is distributed on an "AS IS"
#   basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
#   License for the specific language governing rights and limitations
#   under the License.
#
#   The Original Code is RabbitMQ Management Plugin.
#
#   The Initial Developer of the Original Code is GoPivotal, Inc.
#   Copyright (c) 2007-2020 Pivotal Software, Inc.  All rights reserved.

from __future__ import print_function

import base64
import json
import os
import socket
import ssl
import traceback

import http.client as httplib
import urllib.parse as urlparse
from urllib.parse import quote_plus
from optparse import OptionParser, TitledHelpFormatter

try:
    from signal import signal, SIGPIPE, SIG_DFL

    signal(SIGPIPE, SIG_DFL)
except ImportError:
    pass

import sys


def b64(s):
    return base64.b64encode(s.encode('utf-8')).decode('utf-8')


def eprint(*args, **kwargs):
    print(*args, file=sys.stderr, **kwargs)


if sys.version_info[0] < 2 or (sys.version_info[0] == 2 and sys.version_info[1] < 6):
    eprint("Sorry, rabbitmqadmin requires at least Python 2.6 (2.7.9 when HTTPS is enabled).")
    sys.exit(1)


    def b64(s):
        return base64.b64encode(s.encode('utf-8')).decode('utf-8')

if sys.version_info[0] == 2:
    class ConnectionError(OSError):
        pass


    class ConnectionRefusedError(ConnectionError):
        pass

VERSION = '3.8.3'

LISTABLE = {'connections': {'vhost': False, 'cols': ['name', 'user', 'channels']},
            'channels': {'vhost': False, 'cols': ['name', 'user']},
            'consumers': {'vhost': True},
            'exchanges': {'vhost': True, 'cols': ['name', 'type']},
            'queues': {'vhost': True, 'cols': ['name', 'messages']},
            'bindings': {'vhost': True, 'cols': ['source', 'destination',
                                                 'routing_key']},
            'users': {'vhost': False},
            'vhosts': {'vhost': False, 'cols': ['name', 'messages']},
            'permissions': {'vhost': False},
            'nodes': {'vhost': False, 'cols': ['name', 'type', 'mem_used']},
            'parameters': {'vhost': False, 'json': ['value']},
            'policies': {'vhost': False, 'json': ['definition']},
            'operator_policies': {'vhost': False, 'json': ['definition']},
            'vhost_limits': {'vhost': False, 'json': ['value']}}

SHOWABLE = {'overview': {'vhost': False, 'cols': ['rabbitmq_version',
                                                  'cluster_name',
                                                  'queue_totals.messages',
                                                  'object_totals.queues']}}

PROMOTE_COLUMNS = ['vhost', 'name', 'type',
                   'source', 'destination', 'destination_type', 'routing_key']

URIS = {
    'exchange': '/exchanges/{vhost}/{name}',
    'queue': '/queues/{vhost}/{name}',
    'binding': '/bindings/{vhost}/e/{source}/{destination_char}/{destination}',
    'binding_del': '/bindings/{vhost}/e/{source}/{destination_char}/{destination}/{properties_key}',
    'vhost': '/vhosts/{name}',
    'user': '/users/{name}',
    'permission': '/permissions/{vhost}/{user}',
    'parameter': '/parameters/{component}/{vhost}/{name}',
    'policy': '/policies/{vhost}/{name}',
    'operator_policy': '/operator-policies/{vhost}/{name}',
    'vhost_limit': '/vhost-limits/{vhost}/{name}'
}


def queue_upload_fixup(upload):
    # rabbitmq/rabbitmq-management#761
    #
    # In general, the fixup_upload argument can be used to fixup/change the
    # upload dict after all argument parsing is complete.
    #
    # This simplifies setting the queue type for a new queue by allowing the
    # user to use a queue_type=quorum argument rather than the somewhat confusing
    # arguments='{"x-queue-type":"quorum"}' parameter
    #
    if 'queue_type' in upload:
        queue_type = upload.get('queue_type')
        arguments = upload.get('arguments', {})
        arguments['x-queue-type'] = queue_type
        upload['arguments'] = arguments


DECLARABLE = {
    'exchange': {'mandatory': ['name', 'type'],
                 'json': ['arguments'],
                 'optional': {'auto_delete': 'false', 'durable': 'true',
                              'internal': 'false', 'arguments': {}}},
    'queue': {'mandatory': ['name'],
              'json': ['arguments'],
              'optional': {'auto_delete': 'false', 'durable': 'true',
                           'arguments': {}, 'node': None, 'queue_type': None},
              'fixup_upload': queue_upload_fixup},
    'binding': {'mandatory': ['source', 'destination'],
                'json': ['arguments'],
                'optional': {'destination_type': 'queue',
                             'routing_key': '', 'arguments': {}}},
    'vhost': {'mandatory': ['name'],
              'optional': {'tracing': None}},
    'user': {'mandatory': ['name', ['password', 'password_hash'], 'tags'],
             'optional': {'hashing_algorithm': None}},
    'permission': {'mandatory': ['vhost', 'user', 'configure', 'write', 'read'],
                   'optional': {}},
    'parameter': {'mandatory': ['component', 'name', 'value'],
                  'json': ['value'],
                  'optional': {}},
    # priority has to be converted to an integer
    'policy': {'mandatory': ['name', 'pattern', 'definition'],
               'json': ['definition', 'priority'],
               'optional': {'priority': 0, 'apply-to': None}},
    'operator_policy': {'mandatory': ['name', 'pattern', 'definition'],
                        'json': ['definition', 'priority'],
                        'optional': {'priority': 0, 'apply-to': None}},
    'vhost_limit': {'mandatory': ['vhost', 'name', 'value'],
                    'json': ['value'],
                    'optional': {}},
}

DELETABLE = {
    'exchange': {'mandatory': ['name']},
    'queue': {'mandatory': ['name']},
    'binding': {'mandatory': ['source', 'destination_type', 'destination'],
                'optional': {'properties_key': '~'}},
    'vhost': {'mandatory': ['name']},
    'user': {'mandatory': ['name']},
    'permission': {'mandatory': ['vhost', 'user']},
    'parameter': {'mandatory': ['component', 'name']},
    'policy': {'mandatory': ['name']},
    'operator_policy': {'mandatory': ['name']},
    'vhost_limit': {'mandatory': ['vhost', 'name']}
}

CLOSABLE = {
    'connection': {'mandatory': ['name'],
                   'optional': {},
                   'uri': '/connections/{name}'}
}

PURGABLE = {
    'queue': {'mandatory': ['name'],
              'optional': {},
              'uri': '/queues/{vhost}/{name}/contents'}
}

EXTRA_VERBS = {
    'publish': {'mandatory': ['routing_key'],
                'optional': {'payload': None,
                             'properties': {},
                             'exchange': 'amq.default',
                             'payload_encoding': 'string'},
                'json': ['properties'],
                'uri': '/exchanges/{vhost}/{exchange}/publish'},
    'get': {'mandatory': ['queue'],
            'optional': {'count': '1', 'ackmode': 'ack_requeue_true',
                         'payload_file': None, 'encoding': 'auto'},
            'uri': '/queues/{vhost}/{queue}/get'}
}

for k in DECLARABLE:
    DECLARABLE[k]['uri'] = URIS[k]

for k in DELETABLE:
    DELETABLE[k]['uri'] = URIS[k]
    DELETABLE[k]['optional'] = DELETABLE[k].get('optional', {})
DELETABLE['binding']['uri'] = URIS['binding_del']


def short_usage():
    return "rabbitmqadmin [options] subcommand"


def title(name):
    return "\n%s\n%s\n\n" % (name, '=' * len(name))


def subcommands_usage():
    usage = """Usage
=====
  """ + short_usage() + """

  where subcommand is one of:
""" + title("Display")

    for l in LISTABLE:
        usage += "  list {0} [<column>...]\n".format(l)
    for s in SHOWABLE:
        usage += "  show {0} [<column>...]\n".format(s)
    usage += title("Object Manipulation")
    usage += fmt_usage_stanza(DECLARABLE, 'declare')
    usage += fmt_usage_stanza(DELETABLE, 'delete')
    usage += fmt_usage_stanza(CLOSABLE, 'close')
    usage += fmt_usage_stanza(PURGABLE, 'purge')
    usage += title("Broker Definitions")
    usage += """  export <file>
  import <file>
"""
    usage += title("Publishing and Consuming")
    usage += fmt_usage_stanza(EXTRA_VERBS, '')
    usage += """
  * If payload is not specified on publish, standard input is used

  * If payload_file is not specified on get, the payload will be shown on
    standard output along with the message metadata

  * If payload_file is specified on get, count must not be set
"""
    return usage


def config_usage():
    usage = "Usage\n=====\n" + short_usage()
    usage += "\n" + title("Configuration File")
    usage += """  It is possible to specify a configuration file from the command line.
  Hosts can be configured easily in a configuration file and called
  from the command line.
"""
    usage += title("Example")
    usage += """  # rabbitmqadmin.conf.example START

  [host_normal]
  hostname = localhost
  port = 15672
  username = guest
  password = guest
  declare_vhost = / # Used as default for declare / delete only
  vhost = /         # Used as default for declare / delete / list

  [host_ssl]
  hostname = otherhost
  port = 15672
  username = guest
  password = guest
  ssl = True
  ssl_key_file = /path/to/key.pem
  ssl_cert_file = /path/to/cert.pem

  # rabbitmqadmin.conf.example END
"""
    usage += title("Use")
    usage += """  rabbitmqadmin -c rabbitmqadmin.conf.example -N host_normal ..."""
    return usage


def more_help():
    return """
More Help
=========

For more help use the help subcommand:

  rabbitmqadmin help subcommands  # For a list of available subcommands
  rabbitmqadmin help config       # For help with the configuration file
"""


def fmt_required_flag(val):
    # when one of the options is required, e.g.
    # password vs. password_hash
    if type(val) is list:
        # flag1=... OR flag2=... OR flag3=...
        return "=... OR ".join(val)
    else:
        return val


def fmt_optional_flag(val):
    return val


def fmt_usage_stanza(root, verb):
    def fmt_args(args):
        res = " ".join(["{0}=...".format(fmt_required_flag(a)) for a in args['mandatory']])
        opts = " ".join("{0}=...".format(fmt_optional_flag(o)) for o in args['optional'].keys())
        if opts != "":
            res += " [{0}]".format(opts)
        return res

    text = ""
    if verb != "":
        verb = " " + verb
    for k in root.keys():
        text += " {0} {1} {2}\n".format(verb, k, fmt_args(root[k]))
    return text


def default_config():
    home = os.getenv('USERPROFILE') or os.getenv('HOME')
    if home is not None:
        config_file = home + os.sep + ".rabbitmqadmin.conf"
        if os.path.isfile(config_file):
            return config_file
    return None


def print_version():
    print("rabbitmqadmin {0}".format(VERSION))
    sys.exit(0)


def exit(s):
    eprint("*** {0}\n".format(s))
    sys.exit(1)


class RabbitmqManagement:

    def __init__(self, options):
        default_options = {"hostname": "localhost",
                           "port": "15672",
                           "path_prefix": "",
                           "declare_vhost": "/",
                           "username": "guest",
                           "password": "guest",
                           "ssl": False,
                           "request_timeout": 120,
                           "verbose": True,
                           "format": "table",
                           "depth": 1,
                           "bash_completion": False}

        self.options = default_options
        if options is not None:
            self.options.update(options)

    def get(self, path):
        return self.http("GET", "%s/api%s" % (self.options.get('path_prefix'), path), "")

    def put(self, path, body):
        return self.http("PUT", "%s/api%s" % (self.options.get('path_prefix'), path), body)

    def post(self, path, body):
        return self.http("POST", "%s/api%s" % (self.options.get('path_prefix'), path), body)

    def delete(self, path):
        return self.http("DELETE", "%s/api%s" % (self.options.get('path_prefix'), path), "")

    def __initialize_connection(self, hostname, port):
        if self.options.get('ssl'):
            return self.__initialize_https_connection(hostname, port)
        else:
            return httplib.HTTPConnection(hostname, port, timeout=self.options.get('request_timeout'))

    def __initialize_https_connection(self, hostname, port):
        # Python 2.7.9+
        if hasattr(ssl, 'create_default_context'):
            return httplib.HTTPSConnection(hostname, port, context=self.__initialize_tls_context())
        # Python < 2.7.8, note: those versions still have SSLv3 enabled
        #                       and other limitations. See rabbitmq/rabbitmq-management#225
        else:
            eprint("WARNING: rabbitmqadmin requires Python 2.7.9+ when HTTPS is used.")
            return httplib.HTTPSConnection(hostname, port,
                                           cert_file=self.options.get('ssl_cert_file'),
                                           key_file=self.options.get('ssl_key_file'))

    def __initialize_tls_context(self):
        # Python 2.7.9+ only
        ssl_ctx = ssl.create_default_context(ssl.Purpose.SERVER_AUTH)
        ssl_ctx.options &= ~ssl.OP_NO_SSLv3

        ssl_insecure = self.options.get('ssl_insecure')
        ssl_disable_hostname_verification = ssl_insecure or self.options.get('ssl_disable_hostname_verification')
        # Note: you must set check_hostname prior to verify_mode
        if ssl_disable_hostname_verification:
            ssl_ctx.check_hostname = False
        if ssl_insecure:
            ssl_ctx.verify_mode = ssl.CERT_NONE

        if self.options.get('ssl_key_file'):
            ssl_ctx.load_cert_chain(self.options.get('ssl_cert_file'),
                                    self.options.get('ssl_key_file'))
        if self.options.get('ssl_ca_cert_file'):
            ssl_ctx.load_verify_locations(self.options.get('ssl_ca_cert_file'))
        return ssl_ctx

    def http(self, method, path, body):
        conn = self.__initialize_connection(self.options.get('hostname'), self.options.get('port'))
        auth = (self.options.get('username') + ":" + self.options.get('password'))

        headers = {"Authorization": "Basic " + b64(auth)}
        if body != "":
            headers["Content-Type"] = "application/json"
        try:
            conn.request(method, path, body, headers)
        except ConnectionRefusedError as e:
            exit("Could not connect: {0}".format(e))
        except socket.error as e:
            traceback.print_exc()
            exit("Could not connect: {0}".format(e))
        try:
            resp = conn.getresponse()
        except socket.timeout:
            exit("Timed out getting HTTP response (request timeout: {0} seconds)".format(
                self.options.get('request_timeout')))
        except (KeyboardInterrupt, SystemExit):
            raise
        except (Exception):
            e_fmt = traceback.format_exc()
            exit("Error getting HTTP response:\n\n{0}".format(e_fmt))
        if resp.status == 400:
            exit(json.loads(resp.read())['reason'])
        if resp.status == 401:
            exit("Access refused: {0}".format(path))
        if resp.status == 404:
            exit("Not found: {0}".format(path))
        if resp.status == 301:
            url = urlparse.urlparse(resp.getheader('location'))
            [host, port] = url.netloc.split(':')
            self.options.hostname = host
            self.options.port = int(port)
            return self.http(method, url.path + '?' + url.query, body)
        if resp.status > 400:
            raise Exception("Received response %d %s for path %s\n%s"
                            % (resp.status, resp.reason, path, resp.read()))
        return resp.read().decode('utf-8')

    def verbose(self, string):
        if self.options.get('verbose'):
            print(string)

    def export_def(self, path):
        uri = "/definitions"
        if self.options.get('vhost'):
            uri += "/%s" % quote_plus(self.options.get('vhost'))
        definitions = self.get(uri)
        f = open(path, 'w')
        f.write(definitions)
        f.close()

    def import_def(self, path):
        f = open(path, 'r')
        definitions = f.read()
        f.close()
        uri = "/definitions"
        if self.options.get('vhost'):
            uri += "/%s" % quote_plus(self.options.get('vhost'))
        self.post(uri, definitions)

    def get_queues(self):
        uri = "/queues"
        if self.options.get('vhost'):
            uri += "/%s" % quote_plus(self.options.get('vhost'))
        queues = self.get(uri)
        return json.loads(queues)

    def get_exchanges(self):
        uri = "/exchanges"
        if self.options.get('vhost'):
            uri += "/%s" % quote_plus(self.options.get('vhost'))
        exchanges = self.get(uri)
        return json.loads(exchanges)

    def get_bindings(self):
        uri = "/bindings"
        if self.options.get('vhost'):
            uri += "/%s" % quote_plus(self.options.get('vhost'))
        bindings = self.get(uri)
        return json.loads(bindings)


if __name__ == "__main__":
    rmq = RabbitmqManagement({
        'username': 'guest',
        'password': 'guest'
    })

    # rmq.export_def("/tmp/rmq.json")
    # print(rmq.get_queues())
    # print(rmq.get_exchanges())
    print(rmq.get_bindings())



import re

import pkg_resources

from omt.common import CmdTaskMixin
from omt.core import Resource
from omt.utils import JmxTermUtils


class Bean(Resource, CmdTaskMixin):
    def _run(self):
        pass

    def _completion(self, short_mode=False):
        if self._have_resource_value():
            super()._completion(True)

        else:
            jmx = self.context['jmx'][0] if self.context['jmx'] else ''
            cmd = JmxTermUtils.build_command('open %s && beans' % jmx)
            result = self.run_cmd(cmd, capture_output=True)
            output = result.stdout.decode("utf-8").splitlines()
            output = list(map(lambda x: x.replace(":", "\:"), output))
            self._print_completion(output, True)
            # for line in output:
            #     print(line.replace(":", "\:") + ":" + ' ')

    def info(self):
        jmxterm = pkg_resources.resource_filename(__name__, '../../../lib/jmxterm-1.0.2-uber.jar')
        jmx = self._get_one_resource_value('jmx')
        bean = self._get_one_resource_value()
        bean = bean.replace(" ", "\\ ")
        cmd = 'echo "open %s && bean %s && info"  | java -jar %s -n' % (jmx, bean, jmxterm)
        self.run_cmd(cmd)

    def exec(self):
        if 'completion' in self._get_params():
            jmx = self._get_one_resource_value('jmx')
            bean = self._get_one_resource_value()
            bean = bean.replace(" ", "\\ ")
            short_cmd = "open %s && bean %s && info" % (jmx, bean)
            result = self.run_cmd(JmxTermUtils.build_command(short_cmd), capture_output=True, verbose=False)
            output = result.stdout.decode('utf-8').splitlines()
            for one_attr in (self._list_operation_resources(self._parse_info(output))):
                print(one_attr['method'] + ":" + one_attr['raw_data'])
        else:
            attr_name = ' '.join(self._get_params())
            jmx = self._get_one_resource_value('jmx')
            bean = self._get_one_resource_value()
            bean = bean.replace(" ", "\\ ")
            cmd = "open %s && bean %s && run %s" % (jmx, bean, attr_name)
            # cmd = 'echo "open %s && bean %s && set %s"  | java -jar %s -n' % (jmx, bean, attr_name, jmxterm)
            self.run_cmd(JmxTermUtils.build_command(cmd))

    def get(self):
        if 'completion' in self._get_params():
            jmx = self.context['jmx']
            bean = self._get_resource_values()[0]
            bean = bean.replace(" ", "\\ ")
            short_cmd = "open %s && bean %s && info" % (jmx, bean)
            result = self.run_cmd(JmxTermUtils.build_command(short_cmd), capture_output=True, verbose=False)

            output = result.stdout.decode('utf-8').splitlines()
            for one_attr in (self._list_readable_resources(self._parse_info(output))):
                print(one_attr['attribute'] + ":" + one_attr['raw_data'])
        else:
            attr_name = ' '.join(self._get_params())
            jmxterm = pkg_resources.resource_filename(__name__, '../../../lib/jmxterm-1.0.2-uber.jar')
            jmx = self._get_one_resource_value('jmx')
            bean = self._get_one_resource_value()
            bean = bean.replace(" ", "\\ ")
            cmd = 'echo "open %s && bean %s && get %s"  | java -jar %s -n' % (jmx, bean, attr_name, jmxterm)
            self.run_cmd(cmd)

    def set(self):
        if 'completion' in self._get_params():
            jmx = self._get_one_resource_value('jmx')
            bean = self._get_one_resource_value()
            bean = bean.replace(" ", "\\ ")
            short_cmd = "open %s && bean %s && info" % (jmx, bean)
            result = self.run_cmd(JmxTermUtils.build_command(short_cmd), capture_output=True, verbose=False)
            output = result.stdout.decode('utf-8').splitlines()
            for one_attr in (self._list_writable_resources(self._parse_info(output))):
                print(one_attr['attribute'] + ":" + one_attr['raw_data'])
        else:
            attr_name = ' '.join(self._get_params())
            jmx = self._get_one_resource_value('jmx')
            bean = self._get_one_resource_value()
            bean = bean.replace(" ", "\\ ")
            cmd = "open %s && bean %s && set %s" % (jmx, bean, attr_name)
            self.run_cmd(JmxTermUtils.build_command(cmd))

    def _parse_info(self, result):
        # notification parser is not support right now, should find a example

        '''
        # attributes
        %0   - className (java.lang.String, r)
        %1   - configBaseName (java.lang.String, r)
        %2   - contextClass (java.lang.String, rw)
        %3   - copyXML (boolean, rw)
        %4   - deployXML (boolean, rw)
        %5   - modelerType (java.lang.String, r)
        %6   - unpackWARs (boolean, rw)
        # operations
        %0   - void addServiced(java.lang.String name)
        %1   - void check(java.lang.String name)
        %2   - void checkUndeploy()
        %3   - long getDeploymentTime(java.lang.String name)
        %4   - boolean isDeployed(java.lang.String name)
        %5   - boolean isServiced(java.lang.String name)
        %6   - void manageApp(org.apache.catalina.Context context)
        %7   - void removeServiced(java.lang.String name)
        %8   - void unmanageApp(java.lang.String contextPath)
        '''

        # there's no notifications
        attr_lines = []
        ops_lines = []
        notification_lines = []

        current_section = None
        for line in result:
            if line is None or not line.strip():
                continue

            if '# attributes' in line:
                current_section = 'attributes'
                continue
            if '# operations' in line:
                current_section = 'operations'
                continue

            if '# notifications' in line:
                current_section = 'notifications'
                continue

            if '#there' in line:
                # no content found, e.g. '#there's no notification'
                current_section = None

            if line.startswith('#'):
                # end of section, unknown section
                break

            if current_section == 'attributes':
                attr_lines.append(line)

            if current_section == 'operations':
                ops_lines.append(line)

            if current_section == 'notifications':
                notification_lines.append(line)

            if not current_section:
                continue

        return {
            'attrs': attr_lines,
            'ops': ops_lines,
            'notifications': notification_lines
        }

    def _list_readable_resources(self, parsed_info):
        results = []
        for one_attr_line in parsed_info['attrs']:
            one_result = self._parse_one_attr_line(one_attr_line)
            if one_result['readable']:
                results.append(one_result)
        return results

    def _list_writable_resources(self, parsed_info):
        results = []
        for one_attr_line in parsed_info['attrs']:
            one_result = self._parse_one_attr_line(one_attr_line)
            if one_result['writable']:
                results.append(one_result)
        return results

    def _list_operation_resources(self, parsed_info):
        results = []
        for one_ops_line in parsed_info['ops']:
            one_result = self._parse_one_opeartion_line(one_ops_line)
            results.append(one_result)
        return results

    def _parse_one_attr_line(self, attr):
        # example: '%0   - className (java.lang.String, r)'

        try:
            parsed_result = re.match('.*-(.*)\((.*), (\w+)\)', attr).groups()

            permission = parsed_result[2].strip()

            result = {
                'attribute': parsed_result[0].strip(),
                'param_type': parsed_result[1].strip(),
                'readable': 'r' in permission,
                'writable': 'w' in permission,
                'raw_data': attr
            }

            return result
        except Exception as inst:
            raise Exception("attr: " + str(attr), inst)

    def _parse_one_opeartion_line(self, ops):
        # %0   - void addServiced(java.lang.String name)
        try:
            parsed_result = re.match(".*- ([\w\.]+)(.*)\((.*)\)", ops).groups()

            result = {
                'return': parsed_result[0].strip(),
                'method': parsed_result[1].strip(),
                'param': parsed_result[2].strip(),
                'raw_data': ops
            }
            return result
        except Exception as inst:
            raise Exception("ops: " + str(ops), inst)
            # raise AttributeError(ops)


if __name__ == '__main__':
    result = '''
        # attributes
        %0   - className (java.lang.String, r)
        %1   - configBaseName (java.lang.String, r)
        %2   - contextClass (java.lang.String, rw)
        %3   - copyXML (boolean, rw)
        %4   - deployXML (boolean, rw)
        %5   - modelerType (java.lang.String, r)
        %6   - unpackWARs (boolean, rw)
        # operations
        %0   - void addServiced(java.lang.String name)
        %1   - void check(java.lang.String name)
        %2   - void checkUndeploy()
        %3   - long getDeploymentTime(java.lang.String name)
        %4   - boolean isDeployed(java.lang.String name)
        %5   - boolean isServiced(java.lang.String name)
        %6   - void manageApp(org.apache.catalina.Context context)
        %7   - void removeServiced(java.lang.String name)
        %8   - void unmanageApp(java.lang.String contextPath)
        '''
    #
    # result = (Bean()._parse_info(result.splitlines()))
    # for one_attr in result['attrs']:
    #     print(Bean()._parse_attr(one_attr))
    #
    # for one_op in result['ops']:
    #     print(Bean()._parse_opeartion(one_op))



import os

from omt.core import Resource
import argparse
import json

from omt.core.decorator import filecache
from omt.utils import UrlUtils
from omt.utils.rabbitmq import Management
from omt.config import settings


class Rmq(Resource):
    def __init__(self, context={}, type='web'):
        super().__init__(context, type)
        self.parser = argparse.ArgumentParser()
        self.parser.add_argument('--vhost', nargs='?', help='rabbitmq vhost')
        self.parser.add_argument('url', nargs='?', help='rabbitmq connection string', default='')

    def _description(self):
        return "Rabbitmq Management Command Line Tool"

    def _before_sub_resource(self):
        self.context['common'] = {
            'client': Management(self._build_configuration())
        }

    @filecache(duration=60 * 60 * 24, file=Resource._get_cache_file_name)
    def _completion(self, short_mode=True):
        results = []
        results.append(super()._completion(False))

        if not self._have_resource_value():
            # list rabbitmq connection instance from config file
            config_file_name = os.path.join(settings.RESOURCE_CONFIG_DIR, self.__class__.__name__.lower() + '.json')
            if (os.path.exists(config_file_name)):
                with open(config_file_name) as f:
                    instances = json.load(f)
                    results.extend(
                        self._get_completion([(key, 'instance=' + key) for key, value in instances.items()], False))

        return "\n".join(results)

    def _build_configuration(self):
        args = self.parser.parse_args(self._get_resource_values())
        config = {}
        if args.url:
            if 'http://' in args.url:
                # connection string: http://guest:guest@localhost:15672
                url_utils = UrlUtils(args.url)
                parsed = url_utils.parse()
                config = {
                    'hostname': parsed.hostname if parsed.hostname else 'localhost',
                    'port': parsed.port if parsed.port else 15672,
                    'username': parsed.username if parsed.username else 'guest',
                    'password': parsed.password if parsed.password else 'guest'
                }
            else:
                # parsed as instance, read from config file

                config_file_name = os.path.join(settings.RESOURCE_CONFIG_DIR, self.__class__.__name__.lower() + '.json')

                with open(config_file_name) as f:
                    instances = json.load(f)
                    if args.url in instances:
                        config = instances[args.url]
                pass
        else:
            # no url provided
            pass

        if args.vhost is not None:
            config['vhost'] = args.vhost
            config['declare_vhost'] = args.vhost
        return config


def publish():
    auth = pika.PlainCredentials('guest', 'guest')
    connection = pika.BlockingConnection(pika.ConnectionParameters(host='127.0.0.1', port=5672))
    channel = connection.channel()

    channel.queue_declare(queue='omt_test')

    channel.basic_publish(exchange='',
                          routing_key='omt_test',
                          body='Hello World!')
    print(" [x] Sent 'Hello World!'")
    connection.close()


def consume():
    auth = pika.PlainCredentials('guest', 'guest')
    connection = pika.BlockingConnection(pika.ConnectionParameters(host='127.0.0.1', port=5672))
    channel = connection.channel()

    channel.queue_declare(queue='omt_test')

    def callback(ch, method, properties, body):
        print(" [x] Received %r" % body)

    channel.basic_consume(on_message_callback=callback,
                          queue='omt_test',
                          auto_ack=True)

    print(' [*] Waiting for messages. To exit press CTRL+C')
    channel.start_consuming()


if __name__ == '__main__':
    import sys

    if 'consume' in sys.argv:
        consume()
    elif 'publish' in sys.argv:
        publish()



import json

from omt.utils.rmq_utils import build_admin_params

from omt.common import CompletionMixin
from omt.common.formater import format_list
from omt.core import Resource
import argparse


class Binding(Resource, CompletionMixin):
    def __init__(self, context={}, type='web'):
        super().__init__(context, type)
        self.parser = argparse.ArgumentParser("binding arguments parser")
        self.parser.add_argument('--src', nargs='?', type=str, help='binding source')
        self.parser.add_argument('--dest', nargs='?', type=str, help='binding destination')
        self.parser.add_argument('--type', nargs='?', type=str, help='binding destination type', default='queue')
        self.parser.add_argument('--key', nargs='?', type=str, help='binding property key', default='')
        self.parser.add_argument('--force', nargs='?', type=bool, help='force to delete')

    def _default_columns(self):
        client = self.context['common']['client']
        bindings = client.invoke_list('bindings')
        format_list(bindings, ['source', 'destination', 'routing_key'], 'table')

    def list(self):
        client = self.context['common']['client']
        bindings = client.invoke_list('bindings')
        format_list(bindings)

    def delete(self):
        if 'completion' in self._get_params():
            params = self._get_params()[:-1]
            prompts = []
            if not params:
                # no params, omt rmq binding completion
                self.print_completion(['--src', '--dest'])

            elif params[-1].strip() == '--src':
                # list all exchanges
                client = self.context['common']['client']
                exchanges = json.loads(client.invoke_list('exchanges'))
                results = [(one['name'], "name is %(name)s, type is %(type)s | vhost is %(vhost)s" % one) for one in
                           exchanges]
                self._print_completion(results, short_mode=True)

            elif params[-1].strip() == '--dest':
                # list all queues
                client = self.context['common']['client']
                queues = json.loads(client.invoke_list('queues'))
                results = [(one['name'], "auto_delete is %(auto_delete)s | vhost is %(vhost)s" % one) for one in queues]
                self.print_completion(results, short_mode=True)

            elif params[-1].strip() == '--type':
                # list all queues
                self.print_completion(['exchange', 'queue'], short_mode=True)

            else:
                self.print_completion(['--src', '--dest', '--type', '--key'])
            return

        client = self.context['common']['client']

        args = self.parser.parse_args(self._get_action_params())

        if args.src is None and args.dest is None and args.type is None:
            raise Exception("src,dest and type at least one should be provided")

        client = self.context['common']['client']
        bindings = json.loads(client.invoke_list('bindings'))

        results = []
        for one_binding in bindings:
            if args.src:
                if one_binding['source'] != args.src:
                    continue
            if args.dest:
                if one_binding['destination'] != args.dest:
                    continue
            if args.type:
                if one_binding['destination_type'] != args.type:
                    continue
            if args.key:
                if one_binding['properties_key'] != args.key:
                    continue

            results.append(one_binding)

        if len(results) == 0:
            print('no items found')

        if len(results) > 1 and not args.force:

            print('more than 1 item found:')
            print(results)
            print('please narrow the filter conditions.')
            print('or you can use --force to force to delete all the records')
        else:
            for one in results:
                delete_args = build_admin_params({
                    'source': one['source'],
                    'destination': one['destination'],
                    'destination_type': one['destination_type'],
                    'properties_key': one['properties_key']
                })
                client.invoke_delete('binding', delete_args)

    def declare(self):
        if 'completion' in self._get_params():
            params = self._get_params()[:-1]
            prompts = []
            if not params:
                # no params, omt rmq binding completion
                self.print_completion(['--src', '--dest', '--key'])

            elif params[-1].strip() == '--src':
                # list all exchanges
                client = self.context['common']['client']
                exchanges = json.loads(client.invoke_list('exchanges'))
                results = [(one['name'], "name is %(name)s, type is %(type)s | vhost is %(vhost)s" % one) for one in
                           exchanges]
                self._print_completion(results, short_mode=True)

            elif params[-1].strip() == '--dest':
                # list all queues
                client = self.context['common']['client']
                queues = json.loads(client.invoke_list('queues'))
                results = [(one['name'], "auto_delete is %(auto_delete)s | vhost is %(vhost)s" % one) for one in queues]
                self.print_completion(results, short_mode=True)

            else:
                self.print_completion(['--src', '--dest', '--key'])

            return

        parser = argparse.ArgumentParser('exchange declare arguments')
        parser.add_argument('--src', nargs='?', type=str, help='binding source', required=True)
        parser.add_argument('--dest', nargs='?', type=str, help='binding destination', required=True)
        parser.add_argument('--type', nargs='?', type=str, help='binding destination type', default='queue')
        parser.add_argument('--key', nargs='?', type=str, help='binding property key', default='')

        client = self.context['common']['client']
        args = parser.parse_args(self._get_params())

        client.invoke_declare('binding', build_admin_params({
            'source': args.src,
            'destination': args.dest,
            'routing_key': args.key
        }))



from flask import Flask
app = Flask(__name__)
import omt
import sys
import logging
from omt import utils
import omt
import traceback
import json
from flask import jsonify
@app.route("/zz/<path:path>")
def cmd(path):
    FORMAT = "[%(filename)s:%(lineno)s - %(funcName)s ] %(message)s"
    logging.basicConfig(format=FORMAT)
    logger = logging.getLogger(__name__)
    argv = path.split('/')
    resource_type = argv[0]
    print('########################## resource type:' + resource_type)
    try:
        mod = __import__(".".join(['omt', 'resources', resource_type, resource_type]), fromlist=[resource_type.capitalize()])
        clazz = getattr(mod, resource_type.capitalize())
        context = {
            'all': ['web', *argv],
            'index': 1
        }
        data = clazz(context)._exec()
    
        response = app.response_class(
            response = json.dumps(data),
            mimetype='application/json'
        )
        return response

    except Exception as inst:
        traceback.print_exc()
        logger.error(inst)


# -*- coding: utf-8 -*-
import json
import os
from os.path import exists

from omt.config import settings

from omt.common import CmdTaskMixin
from omt.core import Resource
import pkg_resources

from omt.core.decorator import filecache
from omt.utils import JmxTermUtils


class Jmx(Resource, CmdTaskMixin):
    """
NAME
    jmx - jmx command

SYNOPSIS
    jmx [RESOURCE] action [OPTION]

ACTION LIST

    """

    def _description(self):
        return 'JMX(Java Management Extensions) Tool Set'

    def _run(self):
        jmxterm = pkg_resources.resource_filename(__name__, '../../lib/jmxterm-1.0.2-uber.jar')
        cmd = 'java -jar %s' % jmxterm
        self.run_cmd('echo %s' % cmd)

    def jvms(self):
        jmxterm = pkg_resources.resource_filename(__name__, '../../lib/jmxterm-1.0.2-uber.jar')
        cmd = 'echo jvms | java -jar %s -n' % jmxterm
        self.run_cmd(cmd)

    @filecache(duration=60 * 60 * 24, file=Resource._get_cache_file_name)
    def _completion(self, short_mode=True):
        results = []
        results.append(super()._completion(False))

        if not self._have_resource_value():
            # list rabbitmq connection instance from config file
            config_file_name = os.path.join(settings.CONFIG_DIR, self.__class__.__name__.lower() + '.json')
            if (os.path.exists(config_file_name)):
                with open(config_file_name) as f:
                    instances = json.load(f)
                    results.extend(
                        self._get_completion([(value.get('host') + str(value.get('port')), key) for key, value in instances.items()],
                                             False))

        return "\n".join(results)



import functools
import os
import inspect

from omt.utils.file_utils import make_directory
from collections.abc import Callable


def filecache(duration=None, file: (str, Callable) = '/tmp/cache.txt'):
    def _is_class_method(func):
        spec = inspect.signature(func)
        if len(spec.parameters) > 0:
            if list(spec.parameters.keys())[0] == 'self':
                return True
        return False

    def completion_cache_decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            from datetime import datetime
            if callable(file):
                if _is_class_method(file):
                    cache_file = file(args[0])
                else:
                    cache_file = file()
            else:
                cache_file = file

            cache_is_valid = False

            if not os.path.exists(cache_file):
                cache_is_valid = False
            else:
                if duration is None or duration == -1:
                    cache_is_valid = True
                else:
                    # duration and file all exists
                    os.path.getctime(cache_file)
                    the_duration = datetime.now().timestamp() - os.path.getctime(cache_file)
                    if the_duration > duration:
                        cache_is_valid = False
                    else:
                        cache_is_valid = True

            if cache_is_valid:
                with open(cache_file, 'r') as f:
                    return f.read()

            else:
                # refresh cache
                if os.path.exists(cache_file):
                    os.remove(cache_file)

                make_directory(os.path.dirname(cache_file))
                result = func(*args, **kwargs)
                with open(cache_file, 'w') as f:
                    f.write(result)

                duration_file_name = os.path.join(os.path.dirname(cache_file), 'duration')
                with open(duration_file_name, 'w') as f:
                    f.write("-1" if duration is None else str(duration))

                return result

        return wrapper

    return completion_cache_decorator



#!/usr/bin/env python

import os
import bisect
from datetime import datetime

result = []


def log_analyze(location, keyword):
    results = []
    for root, dir, files in os.walk(location):

        for one_file in files:
            filename = os.path.join(root, one_file)
            items = parse_one_file(filename, keyword)
            if not items:
                continue
            for item in items:
                log_time, line = item
                sort_keys = [one[0] for one in results]
                index = bisect.bisect_left(sort_keys, log_time)
                results.insert(index, (log_time, line, one_file, filename))
        # print(os.path.join(root, files[0]))

    for one_item in results:
        print(one_item[3] + ":")
        print(one_item[1])


def parse_one_file(filename, keyword):
    try:
        with open(filename) as f:
            for line in f.readlines():
                if keyword in line:
                    log_time = None
                    if 'ip=' in line:
                        # for access log
                        log_time = (line.split("ip=", 1)[0]).strip()
                    else:
                        log_time = (line.split("|", 1)[0])

                    the_log_time = datetime.strptime(log_time, "%m-%d-%Y %H:%M:%S.%f")

                    if log_time is None:
                        continue
                    yield (the_log_time, line)
    except Exception as r:
        pass
        # print("can't parse file:" + filename)


if __name__ == '__main__':
    import sys

    location = sys.argv[1]
    keyword = sys.argv[2]
    log_analyze(location, keyword)



import os

header = ['pid', 'username', 'pri', 'nice', 'size', 'res', 'state', 'time', 'wcpu', 'cpu', 'processname', 'session',
          'table', 'client', 'command']


def parse_pg_top(filename):
    with open(filename) as f:
        all_lines = f.readlines()
        start = False
        for one_line in all_lines:

            striped_line = one_line.strip()
            if striped_line.startswith("PID"):
                start = True
                continue

            if not start:
                continue

            result = striped_line.split(None, 14)
            yield dict(zip(header, result))


def parse_pg_top_summary(filename):
    '''Message parser
    last pid: 13501;  load avg:  1.51,  1.97,  2.04;       up 0+02:48:55   11:54:47
467 processes: 13 running, 454 sleeping
CPU states: 41.3% user,  0.0% nice,  4.7% system, 54.0% idle,  0.0% iowait
Memory: 13G used, 18G free, 1292K buffers, 2337M cached
DB activity: 2757 tps,  2 rollbs/s,   3 buffer r/s, 99 hit%,  10369 row r/s,  252 row w/s
DB I/O:     0 reads/s,     0 KB/s,     0 writes/s,     0 KB/s
DB disk: 110.5 GB total, 104.1 GB free (5% used)
Swap: 8192M free
    '''
    with open(filename) as f:
        summary = {}
        all_lines = f.readlines()
        for one_line in all_lines:

            striped_line = one_line.strip()
            if striped_line.startswith("Memory"):
                import re
                a = re.search("(\d*)G free", striped_line)
                summary['free'] = int(a.groups()[0])
                if summary.get('free') is None:
                    print(filename)
                break

        return summary


def is_active(item):
    if item.get('state') == 'run' or (item.get('command') is not None and not item.get('command').endswith('idle')):
        return True


def analyze_result(result):
    summary = {
        'active': 0,
        'lines': 0

    }
    active_count = 0
    lines = 0
    for item in result:
        if is_active(item):
            active_count += 1
        lines = lines + 1

    summary['active'] = active_count
    summary['lines'] = lines
    return summary


if __name__ == '__main__':
    summary = {
        'active_file_name': '',
        'active_count': 0,
        'lines_file_name': '',
        'lines_count': 0,
        'free_file_name': '',
        'free_count': 30
    }
    the_file_name = ''
    count = 0
    for filename in os.listdir("/tmp/pgreport/pgreport"):
        result = parse_pg_top(os.path.join('/tmp/pgreport/pgreport', filename))
        the_result = analyze_result(result)
        active_count = the_result.get('active')
        lines_count = the_result.get('lines')

        if active_count > summary.get('active_count'):
            summary['active_count'] = active_count
            summary['active_file_name'] = filename

        if lines_count > summary.get('lines_count'):
            summary['lines_count'] = lines_count
            summary['lines_file_name'] = filename

        summary_result = parse_pg_top_summary(os.path.join('/tmp/pgreport/pgreport', filename))
        if summary_result.get('free') < summary.get('free_count'):
            summary['free_file_name'] = filename
            summary['free_count'] = summary_result.get('free')

    print(summary)



import functools
import os

import argparse
from omt.common import CmdTaskMixin
from omt.config import settings
from omt.core import simple_completion
from omt.core.decorator import filecache
from omt.core.resource import Resource


class Ssh(Resource, CmdTaskMixin):

    def _description(self):
        return 'SSH(Secure Shell) Smart Tool Set'

    @filecache(duration=60 * 60, file=Resource._get_cache_file_name)
    def _completion(self, short_mode=True):
        results = []
        results.append(super()._completion(True))
        if not self._have_resource_value():
            if not os.path.exists(settings.SSH_CONFIG_FILE):
                return

            ssh_hosts = []
            with open(settings.SSH_CONFIG_FILE) as f:
                for one_line in f.readlines():
                    try:
                        one_line = one_line.strip()
                        if one_line.startswith("Host "):
                            hostname = one_line.replace("Host", "").strip()
                            if hostname:
                                ssh_hosts.append(hostname)
                    except:
                        pass

            results.extend(ssh_hosts)
        return "\n".join(results)

    @simple_completion(['--dry-run'])
    def cache(self):
        parser = argparse.ArgumentParser()
        parser.add_argument('--dry-run', action='store_true')
        args = parser.parse_args(self._get_action_params())
        ssh_host = self._get_one_resource_value()
        cmd = "ssh-copy-id %s" % ssh_host

        if args.dry_run:
            print(cmd)
        else:
            self.run_cmd(cmd)

    def _run(self):
        ssh_host = self._get_one_resource_value()
        cmd = 'ssh %s' % ssh_host
        if '--dry-run' in self._get_resource_values():
            print(cmd)
        else:
            self.run_cmd(cmd)

    @simple_completion(['--dry-run'])
    def exec(self):
        parser = argparse.ArgumentParser()
        parser.add_argument('--dry-run', action='store_true')
        parser.add_argument('cmd', nargs='*')
        args = parser.parse_args(self._get_action_params())

        ssh_host = self._get_one_resource_value()
        cmd = "ssh %s -C '%s'" % (ssh_host, " ".join(args.cmd))
        if args.dry_run:
            print(cmd)
        else:
            self.run_cmd(cmd)

    @simple_completion(['-r', '--local', '--remote', '--dry-run'])
    def upload(self):
        parser = argparse.ArgumentParser()
        parser.add_argument('-r', '--recursive', action='store_true')
        parser.add_argument('--local', nargs='?', help='local files')
        parser.add_argument('--remote', nargs='?', help='remote files')
        parser.add_argument('--dry-run', action='store_true')

        ssh_host = self._get_one_resource_value()

        args = parser.parse_args(self._get_action_params())
        cmd = "scp %s %s %s:%s" % ('-r' if args.recursive else '', args.local, ssh_host, args.remote)
        if args.dry_run:
            print(cmd)
        else:
            self.run_cmd(cmd)

    @simple_completion(['-r', '--local', '--remote', '--recursive', '--dry-run'])
    def download(self):
        parser = argparse.ArgumentParser()
        parser.add_argument('-r', '--recursive', action='store_true')
        parser.add_argument('--local', nargs='?', help='local files')
        parser.add_argument('--remote', nargs='?', help='remote files')
        parser.add_argument('--dry-run', action='store_true')

        ssh_host = self._get_one_resource_value()

        args = parser.parse_args(self._get_action_params())
        cmd = "scp %s %s:%s %s" % ('-r' if args.recursive else '', ssh_host, args.remote, args.local)
        if args.dry_run:
            print(cmd)
        else:
            self.run_cmd(cmd)

    @simple_completion(['--local-port',  '--remote-port', '--remote-host', '--dry-run'])
    def tunnel(self):
        tunnel_template = "ssh -nNT -L %(local_port)s:%(remote_host)s:%(remote_port)s %(bridge)s"
        bridge = self._get_one_resource_value()

        parser = argparse.ArgumentParser()
        parser.add_argument('--local-port', nargs='?', help='local port')
        parser.add_argument('--remote-port', nargs='?', help='remote port')
        parser.add_argument('--remote-host', nargs='?', help='remote host')
        parser.add_argument('--dry-run', action='store_true')

        args = parser.parse_args(self._get_action_params())
        cmd = tunnel_template % {
            'bridge': bridge,
            'local_port': args.local_port,
            'remote_port': args.remote_port,
            'remote_host': args.remote_host,
        }

        if args.dry_run:
            print(cmd)
        else:
            self.run_cmd(cmd)



import os
import sys

from dynaconf import LazySettings


def get_user_home():
    system = sys.platform
    return os.environ['HOME']
    # if system == 'darwin':
    #     return os.environ['HOME']
    # else:
    #     return os.environ["APPDATA"]


def get_home():
    return os.path.join(get_user_home(), '.omt')


settings_file = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'settings.yaml')
settings = LazySettings(
    # PRELOAD_FOR_DYNACONF=[os.path.join(os.environ['HOME'], '.omw', 'config', '*')]
    SETTINGS_FILE_FOR_DYNACONF=settings_file,  # <-- Loaded second (the main file)
    INCLUDES_FOR_DYNACONF=[os.path.join(get_home(), 'config', '*.yaml')]
)



import argparse
import json
import sys
import time

from omt.common.formater import format_list
from omt.core import Resource
import argparse

from omt.core.decorator import filecache
from omt.utils.rmq_utils import build_admin_params


class Exchange(Resource):
    @filecache(duration=60 * 5, file=Resource._get_cache_file_name)
    def _completion(self, short_mode=False):
        super()._completion(True)

        if not self._get_resource_values():
            # resource haven't filled yet
            client = self.context['common']['client']
            exchanges = json.loads(client.invoke_list('exchanges'))
            results = [(one['name'], "name is %(name)s, type is %(type)s | vhost is %(vhost)s" % one) for one in
                       exchanges]
            self._print_completion(results, short_mode=True)

    def list(self):
        client = self.context['common']['client']
        exchanges = client.invoke_list('exchanges')
        format_list(exchanges)

    def delete(self):
        client = self.context['common']['client']
        name = self._get_resource_values()[0]
        client.invoke_delete('exchange', ['name=' + name])

    def declare(self):
        client = self.context['common']['client']
        if not self._have_resource_value():
            raise Exception("no exchange name provided")
        name = self._get_resource_values()[0]
        parser = argparse.ArgumentParser('exchange declare arguments')
        parser.add_argument('--type', nargs='?', default='direct')
        args = parser.parse_args(self._get_params())
        client.invoke_declare('exchange', ['name=' + name, "=".join(['type', args.type])])

    def publish(self):
        client = self.context['common']['client']
        name = self._get_resource_values()[0] if self._have_resource_value() else 'amq.default'
        parser = argparse.ArgumentParser()
        parser.add_argument('--routing-key', nargs='?', help='routing key')
        parser.add_argument('--payload', nargs='?', help='message payload')

        args = parser.parse_args(self._get_params())

        if args.routing_key is None:
            raise Exception("routing-key can't be empty")

        if args.payload is None:
            raise Exception("payload can't be empty")

        params = {
            'exchange': name,
            'routing_key': args.routing_key,
            'payload': args.payload

        }
        client.invoke_publish(build_admin_params(params))

    def listen(self):
        if 'completion' in self._get_params():
            params = self._get_params()[:-1]
            the_completion = ['--queue', '--key', '--ackmode', '--count', '--period']
            if params and params[-1].strip() == '--ackmode':
                the_completion.extend(['ack_requeue_true', 'ack_requeue_false', 'reject_requeue_true', 'reject_requeue_false'])
            self._print_completion(the_completion)
            return

        parser = argparse.ArgumentParser()
        parser.add_argument('--queue', nargs='?', help='queue name', default='_tmp_omt_queue')
        parser.add_argument('--key', nargs='?', help='routing key', default='#')
        parser.add_argument('--ackmode', nargs='?', help='routing key', default='ack_requeue_false')
        parser.add_argument('--count', nargs='?', help='message bulk size', default='1')
        parser.add_argument('--period', nargs='?', help='fetch peroid', default=5)

        args = parser.parse_args(self._get_params())

        queue_name = args.queue
        routing_key = args.key
        exchange_name = self._get_resource_values()[0] if self._have_resource_value() else 'amq.default'

        client = self.context['common']['client']

        try:

            # create temp queue
            client.invoke_declare('queue', ['name=' + queue_name])

            # create binding
            client.invoke_declare('binding', build_admin_params({
                'source': exchange_name,
                'destination': queue_name,
                'routing_key': routing_key
            }))

            while True:
                # fetch message periodically(no amqp support) with ui
                # ['queue=' + queue_name, 'count=' + args.count])
                messages = client.invoke_get(build_admin_params({
                    'queue': queue_name,
                    'count': args.count,
                    'ackmode': args.ackmode
                }))
                if json.loads(messages):
                    format_list(messages)
                time.sleep(args.period)
        except KeyboardInterrupt:
            # delete temp queue and routing key after listen exist
            # purge message
            print('purging message on queue %s' % queue_name)
            client.invoke_purge('queue', ['name=' + queue_name])

            # delete bindings
            print('deleting binding with source %s and destination %s' % (exchange_name, queue_name))
            client.invoke_delete('binding', build_admin_params({
                'source': exchange_name,
                'destination': queue_name,
                'destination_type': 'queue',
                'properties_key': routing_key
            }))

            # delete queue
            print('deleting queue %s' % queue_name)
            client.invoke_delete('queue', ['name=' + queue_name])
            sys.exit(0)



import functools
import os

from omt.core.resource import Resource

__all__ = ['Resource']


def simple_completion(prompts=None):
    def simple_completion_decorator(func):
        @functools.wraps(func)
        def wrapper(self, *args, **kwargs):
            if 'completion' in self._get_action_params():
                self._print_completion(prompts, True)
                return
            else:
                return func(self, *args, **kwargs)

        return wrapper

    return simple_completion_decorator






#!/usr/bin/env python

import logging
import os
import sys
import traceback

import pkg_resources
from omt.core.decorator import filecache

from omt.config import settings


def usage():
    print('''omt resource action params ''')


def main():
    FORMAT = "[%(filename)s:%(lineno)s - %(funcName)s ] %(message)s"
    logging.basicConfig(format=FORMAT)
    logger = logging.getLogger(__name__)
    resource_type = sys.argv[1]

    try:
        mod = __import__(".".join(['omt', 'resources', resource_type, resource_type]),
                         fromlist=[resource_type.capitalize()])
        clazz = getattr(mod, resource_type.capitalize())
        context = {
            'all': sys.argv,
            'index': 1,
            type: 'cmd'
        }
        clazz(context)._exec()
    except Exception as inst:
        traceback.print_exc()
        usage()


if __name__ == '__main__':
    main()



#!/usr/bin/env python

#   The contents of this file are subject to the Mozilla Public License
#   Version 1.1 (the "License"); you may not use this file except in
#   compliance with the License. You may obtain a copy of the License at
#   https://www.mozilla.org/MPL/
#
#   Software distributed under the License is distributed on an "AS IS"
#   basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
#   License for the specific language governing rights and limitations
#   under the License.
#
#   The Original Code is RabbitMQ Management Plugin.
#
#   The Initial Developer of the Original Code is GoPivotal, Inc.
#   Copyright (c) 2007-2020 Pivotal Software, Inc.  All rights reserved.

from __future__ import print_function

from optparse import OptionParser, TitledHelpFormatter

import base64
import json
import os
import socket
import ssl
import traceback

try:
    from signal import signal, SIGPIPE, SIG_DFL

    signal(SIGPIPE, SIG_DFL)
except ImportError:
    pass

import sys


def eprint(*args, **kwargs):
    print(*args, file=sys.stderr, **kwargs)


from configparser import ConfigParser, NoSectionError
import http.client as httplib
import urllib.parse as urlparse
from urllib.parse import quote_plus
from urllib.parse import quote


def b64(s):
    return base64.b64encode(s.encode('utf-8')).decode('utf-8')


class ConnectionRefusedError(ConnectionError):
    pass


VERSION = '3.8.3'

LISTABLE = {'connections': {'vhost': False, 'cols': ['name', 'user', 'channels']},
            'channels': {'vhost': False, 'cols': ['name', 'user']},
            'consumers': {'vhost': True},
            'exchanges': {'vhost': True, 'cols': ['name', 'type']},
            'queues': {'vhost': True, 'cols': ['name', 'messages']},
            'bindings': {'vhost': True, 'cols': ['source', 'destination',
                                                 'routing_key']},
            'users': {'vhost': False},
            'vhosts': {'vhost': False, 'cols': ['name', 'messages']},
            'permissions': {'vhost': False},
            'nodes': {'vhost': False, 'cols': ['name', 'type', 'mem_used']},
            'parameters': {'vhost': False, 'json': ['value']},
            'policies': {'vhost': False, 'json': ['definition']},
            'operator_policies': {'vhost': False, 'json': ['definition']},
            'vhost_limits': {'vhost': False, 'json': ['value']}}

SHOWABLE = {'overview': {'vhost': False, 'cols': ['rabbitmq_version',
                                                  'cluster_name',
                                                  'queue_totals.messages',
                                                  'object_totals.queues']}}

PROMOTE_COLUMNS = ['vhost', 'name', 'type',
                   'source', 'destination', 'destination_type', 'routing_key']

URIS = {
    'exchange': '/exchanges/{vhost}/{name}',
    'queue': '/queues/{vhost}/{name}',
    'binding': '/bindings/{vhost}/e/{source}/{destination_char}/{destination}',
    'binding_del': '/bindings/{vhost}/e/{source}/{destination_char}/{destination}/{properties_key}',
    'vhost': '/vhosts/{name}',
    'user': '/users/{name}',
    'permission': '/permissions/{vhost}/{user}',
    'parameter': '/parameters/{component}/{vhost}/{name}',
    'policy': '/policies/{vhost}/{name}',
    'operator_policy': '/operator-policies/{vhost}/{name}',
    'vhost_limit': '/vhost-limits/{vhost}/{name}'
}


def queue_upload_fixup(upload):
    # rabbitmq/rabbitmq-management#761
    #
    # In general, the fixup_upload argument can be used to fixup/change the
    # upload dict after all argument parsing is complete.
    #
    # This simplifies setting the queue type for a new queue by allowing the
    # user to use a queue_type=quorum argument rather than the somewhat confusing
    # arguments='{"x-queue-type":"quorum"}' parameter
    #
    if 'queue_type' in upload:
        queue_type = upload.get('queue_type')
        arguments = upload.get('arguments', {})
        arguments['x-queue-type'] = queue_type
        upload['arguments'] = arguments


DECLARABLE = {
    'exchange': {'mandatory': ['name', 'type'],
                 'json': ['arguments'],
                 'optional': {'auto_delete': 'false', 'durable': 'true',
                              'internal': 'false', 'arguments': {}}},
    'queue': {'mandatory': ['name'],
              'json': ['arguments'],
              'optional': {'auto_delete': 'false', 'durable': 'true',
                           'arguments': {}, 'node': None, 'queue_type': None},
              'fixup_upload': queue_upload_fixup},
    'binding': {'mandatory': ['source', 'destination'],
                'json': ['arguments'],
                'optional': {'destination_type': 'queue',
                             'routing_key': '', 'arguments': {}}},
    'vhost': {'mandatory': ['name'],
              'optional': {'tracing': None}},
    'user': {'mandatory': ['name', ['password', 'password_hash'], 'tags'],
             'optional': {'hashing_algorithm': None}},
    'permission': {'mandatory': ['vhost', 'user', 'configure', 'write', 'read'],
                   'optional': {}},
    'parameter': {'mandatory': ['component', 'name', 'value'],
                  'json': ['value'],
                  'optional': {}},
    # priority has to be converted to an integer
    'policy': {'mandatory': ['name', 'pattern', 'definition'],
               'json': ['definition', 'priority'],
               'optional': {'priority': 0, 'apply-to': None}},
    'operator_policy': {'mandatory': ['name', 'pattern', 'definition'],
                        'json': ['definition', 'priority'],
                        'optional': {'priority': 0, 'apply-to': None}},
    'vhost_limit': {'mandatory': ['vhost', 'name', 'value'],
                    'json': ['value'],
                    'optional': {}},
}

DELETABLE = {
    'exchange': {'mandatory': ['name']},
    'queue': {'mandatory': ['name']},
    'binding': {'mandatory': ['source', 'destination_type', 'destination'],
                'optional': {'properties_key': '~'}},
    'vhost': {'mandatory': ['name']},
    'user': {'mandatory': ['name']},
    'permission': {'mandatory': ['vhost', 'user']},
    'parameter': {'mandatory': ['component', 'name']},
    'policy': {'mandatory': ['name']},
    'operator_policy': {'mandatory': ['name']},
    'vhost_limit': {'mandatory': ['vhost', 'name']}
}

CLOSABLE = {
    'connection': {'mandatory': ['name'],
                   'optional': {},
                   'uri': '/connections/{name}'}
}

PURGABLE = {
    'queue': {'mandatory': ['name'],
              'optional': {},
              'uri': '/queues/{vhost}/{name}/contents'}
}

EXTRA_VERBS = {
    'publish': {'mandatory': ['routing_key'],
                'optional': {'payload': None,
                             'properties': {},
                             'exchange': 'amq.default',
                             'payload_encoding': 'string'},
                'json': ['properties'],
                'uri': '/exchanges/{vhost}/{exchange}/publish'},
    'get': {'mandatory': ['queue'],
            'optional': {'count': '1', 'ackmode': 'ack_requeue_true',
                         'payload_file': None, 'encoding': 'auto'},
            'uri': '/queues/{vhost}/{queue}/get'}
}

for k in DECLARABLE:
    DECLARABLE[k]['uri'] = URIS[k]

for k in DELETABLE:
    DELETABLE[k]['uri'] = URIS[k]
    DELETABLE[k]['optional'] = DELETABLE[k].get('optional', {})
DELETABLE['binding']['uri'] = URIS['binding_del']


def short_usage():
    return "rabbitmqadmin [options] subcommand"


def title(name):
    return "\n%s\n%s\n\n" % (name, '=' * len(name))


def subcommands_usage():
    usage = """Usage
=====
  """ + short_usage() + """

  where subcommand is one of:
""" + title("Display")

    for l in LISTABLE:
        usage += "  list {0} [<column>...]\n".format(l)
    for s in SHOWABLE:
        usage += "  show {0} [<column>...]\n".format(s)
    usage += title("Object Manipulation")
    usage += fmt_usage_stanza(DECLARABLE, 'declare')
    usage += fmt_usage_stanza(DELETABLE, 'delete')
    usage += fmt_usage_stanza(CLOSABLE, 'close')
    usage += fmt_usage_stanza(PURGABLE, 'purge')
    usage += title("Broker Definitions")
    usage += """  export <file>
  import <file>
"""
    usage += title("Publishing and Consuming")
    usage += fmt_usage_stanza(EXTRA_VERBS, '')
    usage += """
  * If payload is not specified on publish, standard input is used

  * If payload_file is not specified on get, the payload will be shown on
    standard output along with the message metadata

  * If payload_file is specified on get, count must not be set
"""
    return usage


def config_usage():
    usage = "Usage\n=====\n" + short_usage()
    usage += "\n" + title("Configuration File")
    usage += """  It is possible to specify a configuration file from the command line.
  Hosts can be configured easily in a configuration file and called
  from the command line.
"""
    usage += title("Example")
    usage += """  # rabbitmqadmin.conf.example START

  [host_normal]
  hostname = localhost
  port = 15672
  username = guest
  password = guest
  declare_vhost = / # Used as default for declare / delete only
  vhost = /         # Used as default for declare / delete / list

  [host_ssl]
  hostname = otherhost
  port = 15672
  username = guest
  password = guest
  ssl = True
  ssl_key_file = /path/to/key.pem
  ssl_cert_file = /path/to/cert.pem

  # rabbitmqadmin.conf.example END
"""
    usage += title("Use")
    usage += """  rabbitmqadmin -c rabbitmqadmin.conf.example -N host_normal ..."""
    return usage


def more_help():
    return """
More Help
=========

For more help use the help subcommand:

  rabbitmqadmin help subcommands  # For a list of available subcommands
  rabbitmqadmin help config       # For help with the configuration file
"""


def fmt_required_flag(val):
    # when one of the options is required, e.g.
    # password vs. password_hash
    if type(val) is list:
        # flag1=... OR flag2=... OR flag3=...
        return "=... OR ".join(val)
    else:
        return val


def fmt_optional_flag(val):
    return val


def fmt_usage_stanza(root, verb):
    def fmt_args(args):
        res = " ".join(["{0}=...".format(fmt_required_flag(a)) for a in args['mandatory']])
        opts = " ".join("{0}=...".format(fmt_optional_flag(o)) for o in args['optional'].keys())
        if opts != "":
            res += " [{0}]".format(opts)
        return res

    text = ""
    if verb != "":
        verb = " " + verb
    for k in root.keys():
        text += " {0} {1} {2}\n".format(verb, k, fmt_args(root[k]))
    return text


default_options = {"hostname": "localhost",
                   "port": "15672",
                   "path_prefix": "",
                   "declare_vhost": "/",
                   "username": "guest",
                   "password": "guest",
                   "ssl": False,
                   "request_timeout": 120,
                   "verbose": True,
                   "format": "table",
                   "depth": 1,
                   "sort": None,
                   "bash_completion": False}


class MyFormatter(TitledHelpFormatter):
    def format_epilog(self, epilog):
        return epilog


parser = OptionParser(usage=short_usage(),
                      formatter=MyFormatter(),
                      epilog=more_help())


class Obj(object):
    def __init__(self, d):
        self.__dict__ = d


def make_parser():
    def add(*args, **kwargs):
        key = kwargs['dest']
        if key in default_options:
            default = " [default: %s]" % default_options[key]
            kwargs['help'] = kwargs['help'] + default
        parser.add_option(*args, **kwargs)

    add("-c", "--config", dest="config",
        help="configuration file [default: ~/.rabbitmqadmin.conf]",
        metavar="CONFIG")
    add("-N", "--node", dest="node",
        help="node described in the configuration file [default: 'default' only if configuration file is specified]",
        metavar="NODE")
    add("-H", "--host", dest="hostname",
        help="connect to host HOST",
        metavar="HOST")
    add("-P", "--port", dest="port",
        help="connect to port PORT",
        metavar="PORT")
    add("--path-prefix", dest="path_prefix",
        help="use specific URI path prefix for the RabbitMQ HTTP API. /api and operation path will be appended to it. (default: blank string)")
    add("-V", "--vhost", dest="vhost",
        help="connect to vhost VHOST [default: all vhosts for list, '/' for declare]",
        metavar="VHOST")
    add("-u", "--username", dest="username",
        help="connect using username USERNAME",
        metavar="USERNAME")
    add("-p", "--password", dest="password",
        help="connect using password PASSWORD",
        metavar="PASSWORD")
    add("-U", "--base-uri", dest="base_uri",
        help="connect using a base HTTP API URI. /api and operation path will be appended to it. Path will be ignored. --vhost has to be provided separately.",
        metavar="URI")
    add("-q", "--quiet", action="store_false", dest="verbose",
        help="suppress status messages")
    add("-s", "--ssl", action="store_true", dest="ssl",
        help="connect with ssl")
    add("--ssl-key-file", dest="ssl_key_file",
        help="PEM format key file for SSL")
    add("--ssl-cert-file", dest="ssl_cert_file",
        help="PEM format certificate file for SSL")
    add("--ssl-ca-cert-file", dest="ssl_ca_cert_file",
        help="PEM format CA certificate file for SSL")
    add("--ssl-disable-hostname-verification", dest="ssl_disable_hostname_verification",
        help="Disables peer hostname verification", default=False, action="store_true")
    add("-k", "--ssl-insecure", dest="ssl_insecure",
        help="Disables all SSL validations like curl's '-k' argument", default=False, action="store_true")
    add("-t", "--request-timeout", dest="request_timeout",
        help="HTTP request timeout in seconds", type="int")
    add("-f", "--format", dest="format",
        help="format for listing commands - one of [" + ", ".join(FORMATS.keys()) + "]")
    add("-S", "--sort", dest="sort", help="sort key for listing queries")
    add("-R", "--sort-reverse", action="store_true", dest="sort_reverse",
        help="reverse the sort order")
    add("-d", "--depth", dest="depth",
        help="maximum depth to recurse for listing tables")
    add("--bash-completion", action="store_true",
        dest="bash_completion",
        help="Print bash completion script")
    add("--version", action="store_true",
        dest="version",
        help="Display version and exit")


def default_config():
    home = os.getenv('USERPROFILE') or os.getenv('HOME')
    if home is not None:
        config_file = home + os.sep + ".rabbitmqadmin.conf"
        if os.path.isfile(config_file):
            return config_file
    return None


def make_configuration():
    make_parser()
    (options, args) = parser.parse_args()
    setattr(options, "declare_vhost", None)
    if options.version:
        print_version()
    if options.config is None:
        config_file = default_config()
        if config_file is not None:
            setattr(options, "config", config_file)
    else:
        if not os.path.isfile(options.config):
            assert_usage(False, "Could not read config file '%s'" % options.config)

    if options.node is None and options.config:
        options.node = "default"
    else:
        options.node = options.node
    for (key, val) in default_options.items():
        if getattr(options, key) is None:
            setattr(options, key, val)

    if options.config is not None:
        config = ConfigParser()
        try:
            config.read(options.config)
            new_conf = dict(config.items(options.node))
        except NoSectionError as error:
            if options.node == "default":
                pass
            else:
                msg = "Could not read section '%s' in config file '%s':\n   %s" % (options.node, options.config, error)
                assert_usage(False, msg)
        else:
            for key, val in new_conf.items():
                if key == 'ssl':
                    setattr(options, key, val == "True")
                else:
                    setattr(options, key, val)

    # if --base-uri is passed, set connection parameters from it
    if options.base_uri is not None:
        u = urlparse.urlparse(options.base_uri)
        for key in ["hostname", "port", "username", "password"]:
            if getattr(u, key) is not None:
                setattr(options, key, getattr(u, key))

        if u.path is not None and (u.path != "") and (u.path != "/"):
            eprint("WARNING: path in --base-uri is ignored. Please specify --vhost and/or --path-prefix separately.\n")

    return (options, args)


def assert_usage(expr, error):
    if not expr:
        eprint("\nERROR: {0}\n".format(error))
        eprint("{0} --help for help\n".format(os.path.basename(sys.argv[0])))
        sys.exit(1)


def print_version():
    print("rabbitmqadmin {0}".format(VERSION))
    sys.exit(0)


def column_sort_key(col):
    if col in PROMOTE_COLUMNS:
        return (1, PROMOTE_COLUMNS.index(col))
    else:
        return (2, col)


def main():
    (options, args) = make_configuration()
    if options.bash_completion:
        print_bash_completion()
        sys.exit(0)
    assert_usage(len(args) > 0, 'Action not specified')
    mgmt = Management(options)
    mode = "invoke_" + args[0]
    assert_usage(hasattr(mgmt, mode),
                 'Action {0} not understood'.format(args[0]))
    method = getattr(mgmt, "invoke_%s" % args[0])
    method(args[1:])


def die(s):
    eprint("*** {0}\n".format(s))
    sys.exit(1)


def maybe_utf8(s):
    if isinstance(s, int):
        # s can be also an int for ex messages count
        return str(s)
    if isinstance(s, float):
        # s can be also a float for message rate
        return str(s)
    if sys.version_info[0] == 3:
        # It will have an encoding, which Python will respect
        return s
    else:
        # It won't have an encoding, and Python will pick ASCII by default
        return s.encode('utf-8')


class Management:
    def __init__(self, options):
        the_default_options = {"hostname": "localhost",
                               "port": "15672",
                               "path_prefix": "",
                               "declare_vhost": "/",
                               "username": "guest",
                               "password": "guest",
                               "ssl": False,
                               "request_timeout": 120,
                               "verbose": True,
                               "format": "table",
                               "depth": 1,
                                "vhost": "/",
                               "sort": None,
                               "sort_reverse": None,
                               "bash_completion": False
                               }
        tmp_options = the_default_options
        tmp_options.update(options)
        self.options = Obj(tmp_options)
        # self.args = args

    def get(self, path):
        return self.http("GET", "%s/api%s" % (self.options.path_prefix, path), "")

    def put(self, path, body):
        return self.http("PUT", "%s/api%s" % (self.options.path_prefix, path), body)

    def post(self, path, body):
        return self.http("POST", "%s/api%s" % (self.options.path_prefix, path), body)

    def delete(self, path):
        return self.http("DELETE", "%s/api%s" % (self.options.path_prefix, path), "")

    def __initialize_connection(self, hostname, port):
        if self.options.ssl:
            return self.__initialize_https_connection(hostname, port)
        else:
            return httplib.HTTPConnection(hostname, port, timeout=self.options.request_timeout)

    def __initialize_https_connection(self, hostname, port):
        # Python 2.7.9+
        if hasattr(ssl, 'create_default_context'):
            return httplib.HTTPSConnection(hostname, port, context=self.__initialize_tls_context())
        # Python < 2.7.8, note: those versions still have SSLv3 enabled
        #                       and other limitations. See rabbitmq/rabbitmq-management#225
        else:
            eprint("WARNING: rabbitmqadmin requires Python 2.7.9+ when HTTPS is used.")
            return httplib.HTTPSConnection(hostname, port,
                                           cert_file=self.options.ssl_cert_file,
                                           key_file=self.options.ssl_key_file)

    def __initialize_tls_context(self):
        # Python 2.7.9+ only
        ssl_ctx = ssl.create_default_context(ssl.Purpose.SERVER_AUTH)
        ssl_ctx.options &= ~ssl.OP_NO_SSLv3

        ssl_insecure = self.options.ssl_insecure
        ssl_disable_hostname_verification = ssl_insecure or self.options.ssl_disable_hostname_verification
        # Note: you must set check_hostname prior to verify_mode
        if ssl_disable_hostname_verification:
            ssl_ctx.check_hostname = False
        if ssl_insecure:
            ssl_ctx.verify_mode = ssl.CERT_NONE

        if self.options.ssl_key_file:
            ssl_ctx.load_cert_chain(self.options.ssl_cert_file,
                                    self.options.ssl_key_file)
        if self.options.ssl_ca_cert_file:
            ssl_ctx.load_verify_locations(self.options.ssl_ca_cert_file)
        return ssl_ctx

    def http(self, method, path, body):
        conn = self.__initialize_connection(self.options.hostname, self.options.port)
        auth = (self.options.username + ":" + self.options.password)

        headers = {"Authorization": "Basic " + b64(auth)}
        if body != "":
            headers["Content-Type"] = "application/json"
        try:
            conn.request(method, path, body, headers)
        except ConnectionRefusedError as e:
            die("Could not connect: {0}".format(e))
        except socket.error as e:
            traceback.print_exc()
            die("Could not connect: {0}".format(e))
        try:
            resp = conn.getresponse()
        except socket.timeout:
            die("Timed out getting HTTP response (request timeout: {0} seconds)".format(
                self.options.request_timeout))
        except (KeyboardInterrupt, SystemExit):
            raise
        except (Exception):
            e_fmt = traceback.format_exc()
            die("Error getting HTTP response:\n\n{0}".format(e_fmt))
        if resp.status == 400:
            die(json.loads(resp.read())['reason'])
        if resp.status == 401:
            die("Access refused: {0}".format(path))
        if resp.status == 404:
            die("Not found: {0}".format(path))
        if resp.status == 301:
            url = urlparse.urlparse(resp.getheader('location'))
            [host, port] = url.netloc.split(':')
            self.options.hostname = host
            self.options.port = int(port)
            return self.http(method, url.path + '?' + url.query, body)
        if resp.status > 400:
            raise Exception("Received response %d %s for path %s\n%s"
                            % (resp.status, resp.reason, path, resp.read()))
        return resp.read().decode('utf-8')

    def verbose(self, string):
        if self.options.verbose:
            print(string)

    def get_arg(self):
        assert_usage(len(self.args) == 1, 'Exactly one argument required')
        return self.args[0]

    def use_cols(self):
        # Deliberately do not cast to int here; we only care about the
        # default, not explicit setting.
        return self.options.depth == 1 and ('json' not in self.options.format)

    def invoke_help(self, args=None):
        if len(args) == 0:
            parser.print_help()
        else:
            help_cmd = self.get_arg()
            if help_cmd == 'subcommands':
                usage = subcommands_usage()
            elif help_cmd == 'config':
                usage = config_usage()
            else:
                assert_usage(False, """help topic must be one of:
  subcommands
  config""")
            print(usage)
        sys.exit(0)

    def invoke_publish(self, args=None):
        (uri, upload) = self.parse_args(args, EXTRA_VERBS['publish'])
        if 'payload' not in upload:
            data = sys.stdin.read()
            upload['payload'] = b64(data)
            upload['payload_encoding'] = 'base64'
        resp = json.loads(self.post(uri, json.dumps(upload)))
        if resp['routed']:
            self.verbose("Message published")
        else:
            self.verbose("Message published but NOT routed")

    def invoke_get(self, args=None):
        (uri, upload) = self.parse_args(args, EXTRA_VERBS['get'])
        payload_file = 'payload_file' in upload and upload['payload_file'] or None
        assert_usage(not payload_file or upload['count'] == '1',
                     'Cannot get multiple messages using payload_file')
        result = self.post(uri, json.dumps(upload))
        return result
        # if payload_file:
        #     write_payload_file(payload_file, result)
        #     columns = ['routing_key', 'exchange', 'message_count',
        #                'payload_bytes', 'redelivered']
        #     format_list(result, columns, {}, self.options)
        # else:
        #     format_list(result, [], {}, self.options)

    def invoke_export(self):
        path = self.get_arg()
        uri = "/definitions"
        if self.options.vhost:
            uri += "/%s" % quote_plus(self.options.vhost)
        definitions = self.get(uri)
        f = open(path, 'w')
        f.write(definitions)
        f.close()
        self.verbose("Exported definitions for %s to \"%s\""
                     % (self.options.hostname, path))

    def invoke_import(self):
        path = self.get_arg()
        f = open(path, 'r')
        definitions = f.read()
        f.close()
        uri = "/definitions"
        if self.options.vhost:
            uri += "/%s" % quote_plus(self.options.vhost)
        self.post(uri, definitions)
        self.verbose(
            "Uploaded definitions from \"%s\" to %s. The import process may take some time. Consult server logs to track progress."
            % (self.options.hostname, path))

    def invoke_list(self, obj_type, args=None):
        (uri, obj_info, cols) = self.list_show_uri(LISTABLE, 'list', obj_type, args)
        return self.get(uri)
        #format_list(self.get(uri), cols, obj_info, self.options)

    def invoke_show(self):
        (uri, obj_info, cols) = self.list_show_uri(SHOWABLE, 'show')
        format_list('[{0}]'.format(self.get(uri)), cols, obj_info, self.options)

    def _list_path_for_obj_type(self, obj_type):
        # This returns a URL path for given object type, e.g.
        # replaces underscores in command names with
        # dashes that HTTP API endpoints use
        return obj_type.replace("_", "-")

    def list_show_uri(self, obj_types, verb, obj_type, args):
        if args is None:
            args = []
        assert_usage(obj_type in obj_types,
                     "Don't know how to {0} {1}".format(verb, obj_type))
        obj_info = obj_types[obj_type]
        uri = "/%s" % self._list_path_for_obj_type(obj_type)
        query = []
        if obj_info['vhost'] and self.options.vhost:
            uri += "/%s" % quote_plus(self.options.vhost)
        cols = args
        if cols == [] and 'cols' in obj_info and self.use_cols():
            cols = obj_info['cols']
        if cols != []:
            pass
            # query.append("columns=" + ",".join(cols))
        sort = self.options.sort
        if sort:
            query.append("sort=" + sort)
        if self.options.sort_reverse:
            query.append("sort_reverse=true")
        query = "&".join(query)
        if query != "":
            uri += "?" + query
        return (uri, obj_info, cols)

    def invoke_declare(self, obj_type, args):
        (obj_type, uri, upload) = self.declare_delete_parse(DECLARABLE, obj_type, args)
        if obj_type == 'binding':
            self.post(uri, json.dumps(upload))
        else:
            self.put(uri, json.dumps(upload))
        self.verbose("{0} declared".format(obj_type))

    def invoke_delete(self, obj_type, args):
        (obj_type, uri, upload) = self.declare_delete_parse(DELETABLE, obj_type, args)
        self.delete(uri)
        self.verbose("{0} deleted".format(obj_type))

    def invoke_close(self):
        (obj_type, uri, upload) = self.declare_delete_parse(CLOSABLE)
        self.delete(uri)
        self.verbose("{0} closed".format(obj_type))

    def invoke_purge(self, obj_type, args):
        (obj_type, uri, upload) = self.declare_delete_parse(PURGABLE, obj_type, args)
        self.delete(uri)
        self.verbose("{0} purged".format(obj_type))

    def declare_delete_parse(self, root, obj_type, args):
        #assert_usage(len(self.args) > 0, 'Type not specified')
        # obj_type = self.args[0]
        assert_usage(obj_type in root,
                     'Type {0} not recognised'.format(obj_type))
        obj = root[obj_type]
        (uri, upload) = self.parse_args(args, obj)
        return (obj_type, uri, upload)

    def assert_mandatory_keys(self, mandatory, upload):
        for m in mandatory:
            if type(m) is list:
                a_set = set(m)
                b_set = set(upload.keys())
                assert_usage((a_set & b_set),
                             'one of mandatory arguments "{0}" is required'.format(m))
            else:
                assert_usage(m in upload.keys(),
                             'mandatory argument "{0}" is required'.format(m))

    def parse_args(self, args, obj):
        mandatory = obj['mandatory']
        optional = obj['optional']
        uri_template = obj['uri']
        upload = {}
        for k in optional.keys():
            if optional[k] is not None:
                upload[k] = optional[k]
        for arg in args:
            assert_usage("=" in arg,
                         'Argument "{0}" not in the name=value format'.format(arg))
            (name, value) = arg.split("=", 1)
            # flatten the list of mandatory keys
            mandatory_keys = []
            for key in mandatory:
                if type(key) is list:
                    for subkey in key:
                        mandatory_keys.append(subkey)
                else:
                    mandatory_keys.append(key)

            assert_usage(name in mandatory_keys or name in optional.keys(),
                         'Argument "{0}" is not recognised'.format(name))

            if 'json' in obj and name in obj['json']:
                upload[name] = self.parse_json(value)
            else:
                upload[name] = value
        self.assert_mandatory_keys(mandatory, upload)
        if 'vhost' not in mandatory:
            upload['vhost'] = self.options.vhost or self.options.declare_vhost
        uri_args = {}
        for k in upload:
            v = upload[k]
            if v is not None and isinstance(v, (str, bytes)):
                uri_args[k] = quote(v, '')
                if k == 'destination_type':
                    uri_args['destination_char'] = v[0]
        uri = uri_template.format(**uri_args)
        if 'fixup_upload' in obj:
            fixup = obj['fixup_upload']
            fixup(upload)
        return (uri, upload)

    def parse_json(self, text):
        try:
            return json.loads(text)
        except ValueError:
            eprint("ERROR: Could not parse JSON:\n  {0}".format(text))
            sys.exit(1)


def format_list(json_list, columns, args, options):
    format = options.format
    formatter = None
    if format == "raw_json":
        print(json_list)
        return
    elif format == "pretty_json":
        enc = json.JSONEncoder(False, False, True, True, True, 2)
        print(enc.encode(json.loads(json_list)))
        return
    else:
        formatter = FORMATS[format]
    assert_usage(formatter is not None,
                 "Format {0} not recognised".format(format))
    formatter_instance = formatter(columns, args, options)
    formatter_instance.display(json_list)


class Lister:
    def verbose(self, string):
        if self.options.verbose:
            print(string)

    def display(self, json_list):
        depth = sys.maxsize
        if len(self.columns) == 0:
            depth = int(self.options.depth)
        (columns, table) = self.list_to_table(json.loads(json_list), depth)
        if len(table) > 0:
            self.display_list(columns, table)
        else:
            self.verbose("No items")

    def list_to_table(self, items, max_depth):
        columns = {}
        column_ix = {}
        row = None
        table = []

        def add(prefix, depth, item, fun):
            for key in item:
                column = prefix == '' and key or (prefix + '.' + key)
                subitem = item[key]
                if type(subitem) == dict:
                    if 'json' in self.obj_info and key in self.obj_info['json']:
                        fun(column, json.dumps(subitem))
                    else:
                        if depth < max_depth:
                            add(column, depth + 1, subitem, fun)
                elif type(subitem) == list:
                    # The first branch has mirrors in queues in
                    # mind (which come out looking decent); the second
                    # one has applications in nodes (which look less
                    # so, but what would look good?).
                    if [x for x in subitem if type(x) != str] == []:
                        serialised = " ".join(subitem)
                    else:
                        serialised = json.dumps(subitem)
                    fun(column, serialised)
                else:
                    fun(column, subitem)

        def add_to_columns(col, val):
            columns[col] = True

        def add_to_row(col, val):
            if col in column_ix:
                if val is not None:
                    row[column_ix[col]] = maybe_utf8(val)
                else:
                    row[column_ix[col]] = None

        if len(self.columns) == 0:
            for item in items:
                add('', 1, item, add_to_columns)
            columns = list(columns.keys())
            columns.sort(key=column_sort_key)
        else:
            columns = self.columns

        for i in range(0, len(columns)):
            column_ix[columns[i]] = i
        for item in items:
            row = len(columns) * ['']
            add('', 1, item, add_to_row)
            table.append(row)

        return (columns, table)


class TSVList(Lister):
    def __init__(self, columns, obj_info, options):
        self.columns = columns
        self.obj_info = obj_info
        self.options = options

    def display_list(self, columns, table):
        head = "\t".join(columns)
        self.verbose(head)

        for row in table:
            line = "\t".join(row)
            print(line)


class LongList(Lister):
    def __init__(self, columns, obj_info, options):
        self.columns = columns
        self.obj_info = obj_info
        self.options = options

    def display_list(self, columns, table):
        sep = "\n" + "-" * 80 + "\n"
        max_width = 0
        for col in columns:
            max_width = max(max_width, len(col))
        fmt = "{0:>" + str(max_width) + "}: {1}"
        print(sep)
        for i in range(0, len(table)):
            for j in range(0, len(columns)):
                print(fmt.format(columns[j], table[i][j]))
            print(sep)


class TableList(Lister):
    def __init__(self, columns, obj_info, options):
        self.columns = columns
        self.obj_info = obj_info
        self.options = options

    def display_list(self, columns, table):
        total = [columns]
        total.extend(table)
        self.ascii_table(total)

    def ascii_table(self, rows):
        col_widths = [0] * len(rows[0])
        for i in range(0, len(rows[0])):
            for j in range(0, len(rows)):
                col_widths[i] = max(col_widths[i], len(rows[j][i]))
        self.ascii_bar(col_widths)
        self.ascii_row(col_widths, rows[0], "^")
        self.ascii_bar(col_widths)
        for row in rows[1:]:
            self.ascii_row(col_widths, row, "<")
        self.ascii_bar(col_widths)

    def ascii_row(self, col_widths, row, align):
        txt = "|"
        for i in range(0, len(col_widths)):
            fmt = " {0:" + align + str(col_widths[i]) + "} "
            txt += fmt.format(row[i]) + "|"
        print(txt)

    def ascii_bar(self, col_widths):
        txt = "+"
        for w in col_widths:
            txt += ("-" * (w + 2)) + "+"
        print(txt)


class KeyValueList(Lister):
    def __init__(self, columns, obj_info, options):
        self.columns = columns
        self.obj_info = obj_info
        self.options = options

    def display_list(self, columns, table):
        for i in range(0, len(table)):
            row = []
            for j in range(0, len(columns)):
                row.append("{0}=\"{1}\"".format(columns[j], table[i][j]))
            print(" ".join(row))


# TODO handle spaces etc in completable names
class BashList(Lister):
    def __init__(self, columns, obj_info, options):
        self.columns = columns
        self.obj_info = obj_info
        self.options = options

    def display_list(self, columns, table):
        ix = None
        for i in range(0, len(columns)):
            if columns[i] == 'name':
                ix = i
        if ix is not None:
            res = []
            for row in table:
                res.append(row[ix])
            print(" ".join(res))


FORMATS = {
    # Special cased
    'raw_json': None,
    # Ditto
    'pretty_json': None,
    'tsv': TSVList,
    'long': LongList,
    'table': TableList,
    'kvp': KeyValueList,
    'bash': BashList
}


def write_payload_file(payload_file, json_list):
    result = json.loads(json_list)[0]
    payload = result['payload']
    payload_encoding = result['payload_encoding']
    f = open(payload_file, 'w')
    if payload_encoding == 'base64':
        data = base64.b64decode(payload)
    else:
        data = payload
    f.write(data)
    f.close()


def print_bash_completion():
    script = """# This is a bash completion script for rabbitmqadmin.
# Redirect it to a file, then source it or copy it to /etc/bash_completion.d
# to get tab completion. rabbitmqadmin must be on your PATH for this to work.
_rabbitmqadmin()
{
    local cur prev opts base
    COMPREPLY=()
    cur="${COMP_WORDS[COMP_CWORD]}"
    prev="${COMP_WORDS[COMP_CWORD-1]}"

    opts="list show declare delete close purge import export get publish help"
    fargs="--help --host --port --vhost --username --password --format --depth --sort --sort-reverse"

    case "${prev}" in
        list)
            COMPREPLY=( $(compgen -W '""" + " ".join(LISTABLE) + """' -- ${cur}) )
            return 0
            ;;
        show)
            COMPREPLY=( $(compgen -W '""" + " ".join(SHOWABLE) + """' -- ${cur}) )
            return 0
            ;;
        declare)
            COMPREPLY=( $(compgen -W '""" + " ".join(DECLARABLE.keys()) + """' -- ${cur}) )
            return 0
            ;;
        delete)
            COMPREPLY=( $(compgen -W '""" + " ".join(DELETABLE.keys()) + """' -- ${cur}) )
            return 0
            ;;
        close)
            COMPREPLY=( $(compgen -W '""" + " ".join(CLOSABLE.keys()) + """' -- ${cur}) )
            return 0
            ;;
        purge)
            COMPREPLY=( $(compgen -W '""" + " ".join(PURGABLE.keys()) + """' -- ${cur}) )
            return 0
            ;;
        export)
            COMPREPLY=( $(compgen -f ${cur}) )
            return 0
            ;;
        import)
            COMPREPLY=( $(compgen -f ${cur}) )
            return 0
            ;;
        help)
            opts="subcommands config"
            COMPREPLY=( $(compgen -W "${opts}"  -- ${cur}) )
            return 0
            ;;
        -H)
            COMPREPLY=( $(compgen -A hostname ${cur}) )
            return 0
            ;;
        --host)
            COMPREPLY=( $(compgen -A hostname ${cur}) )
            return 0
            ;;
        -V)
            opts="$(rabbitmqadmin -q -f bash list vhosts)"
            COMPREPLY=( $(compgen -W "${opts}"  -- ${cur}) )
            return 0
            ;;
        --vhost)
            opts="$(rabbitmqadmin -q -f bash list vhosts)"
            COMPREPLY=( $(compgen -W "${opts}"  -- ${cur}) )
            return 0
            ;;
        -u)
            opts="$(rabbitmqadmin -q -f bash list users)"
            COMPREPLY=( $(compgen -W "${opts}"  -- ${cur}) )
            return 0
            ;;
        --username)
            opts="$(rabbitmqadmin -q -f bash list users)"
            COMPREPLY=( $(compgen -W "${opts}"  -- ${cur}) )
            return 0
            ;;
        -f)
            COMPREPLY=( $(compgen -W \"""" + " ".join(FORMATS.keys()) + """\"  -- ${cur}) )
            return 0
            ;;
        --format)
            COMPREPLY=( $(compgen -W \"""" + " ".join(FORMATS.keys()) + """\"  -- ${cur}) )
            return 0
            ;;

"""
    for l in LISTABLE:
        key = l[0:len(l) - 1]
        script += "        " + key + """)
            opts="$(rabbitmqadmin -q -f bash list """ + l + """)"
            COMPREPLY=( $(compgen -W "${opts}"  -- ${cur}) )
            return 0
            ;;
"""
    script += """        *)
        ;;
    esac

   COMPREPLY=($(compgen -W "${opts} ${fargs}" -- ${cur}))
   return 0
}
complete -F _rabbitmqadmin rabbitmqadmin
"""
    print(script)


if __name__ == "__main__":
    main()



import argparse
import json
import sys
import time

from omt.core.decorator import filecache
from omt.utils.rmq_utils import build_admin_params

from omt.common import CompletionMixin
from omt.common.formater import format_list
from omt.core import Resource


class Queue(Resource, CompletionMixin):
    @filecache(duration=60 * 5, file=Resource._get_cache_file_name)
    def _completion(self, short_mode=True):
        super()._completion(short_mode)

        if not self._have_resource_value():
            # completions for queue name
            client = self.context['common']['client']
            queues = json.loads(client.invoke_list('queues'))
            results = [(one['name'], "auto_delete is %(auto_delete)s | vhost is %(vhost)s" % one) for one in queues]
            self.print_completion(results, short_mode=True)

    def list(self):
        client = self.context['common']['client']
        queues = client.invoke_list('queues')
        format_list(queues)

    def get(self):
        parser = argparse.ArgumentParser()
        parser.add_argument('--count', nargs="?", help='message count', type=str, default='100')
        args = parser.parse_args(self._get_params())
        client = self.context['common']['client']
        queue_name = self._get_resource_values()[0]
        messages = client.invoke_get(['queue=' + queue_name, 'count=' + args.count])
        format_list(messages)

    def delete(self):
        client = self.context['common']['client']
        queue_name = self._get_resource_values()[0]
        client.invoke_delete('queue', ['name=' + queue_name])

    def declare(self):
        # parser = argparse.ArgumentParser('exchange declare arguments')
        # parser.add_argument('--type', nargs='?', default='direct')

        client = self.context['common']['client']
        name = self._get_resource_values()[0]
        # args = parser.parse_args(self._get_params())

        client.invoke_declare('queue', ['name=' + name])

    def publish(self):
        '''Message will be published to the default exchange(amq.default) with routing key queue_name, routing it to this queue.'''
        client = self.context['common']['client']
        if not self._have_resource_value():
            raise Exception("no queue name provided")
        name = self._get_resource_values()[0]
        parser = argparse.ArgumentParser()
        parser.add_argument('--payload', nargs='?', help='message payload')

        args = parser.parse_args(self._get_params())

        if args.payload is None:
            raise Exception("payload can't be empty")

        params = {
            'exchange': 'amq.default',
            'routing_key': name,
            'payload': args.payload

        }
        client.invoke_publish(build_admin_params(params))

    def purge(self):
        client = self.context['common']['client']
        if not self._have_resource_value():
            raise Exception("no queue name provided")
        name = self._get_resource_values()[0]
        client.invoke_purge('queue', ['name=' + name])

    def listen(self):
        if 'completion' in self._get_params():
            params = self._get_params()[:-1]
            the_completion = ['--ackmode', '--count', '--period']
            if params and params[-1].strip() == '--ackmode':
                the_completion.extend(['ack_requeue_true', 'ack_requeue_false', 'reject_requeue_true', 'reject_requeue_false'])
            self._print_completion(the_completion)
            return

        parser = argparse.ArgumentParser()
        parser.add_argument('--ackmode', nargs='?', help='routing key', default='ack_requeue_false')
        parser.add_argument('--count', nargs='?', help='message bulk size', default='1')
        parser.add_argument('--period', nargs='?', help='fetch peroid', default=5)

        args = parser.parse_args(self._get_params())

        if not self._get_resource_values():
            raise Exception("no queue provided")

        queue_name = self._get_resource_values()[0]

        client = self.context['common']['client']

        try:
            print('star listening message on queue %s, press Ctrl-C to stop' % queue_name)
            while True:
                # fetch message periodically(no amqp support) with ui
                # ['queue=' + queue_name, 'count=' + args.count])
                messages = client.invoke_get(build_admin_params({
                    'queue': queue_name,
                    'count': args.count,
                    'ackmode': args.ackmode
                }))
                if json.loads(messages):
                    format_list(messages)
                time.sleep(args.period)
        except KeyboardInterrupt:
            print('stop listening')
            sys.exit(0)



from dynaconf import settings
import subprocess
import os


class CmdTaskMixin:

    def run_cmd(self, cmd, cwd=None, env=None, block=True, capture_output=False, verbose=True, *args, **kwargs):
        the_env = env if not env else os.environ
        try:
            cwd = cwd if cwd is None else cwd.replace("\\", "/")

            if verbose:
                print("cmd: %s, cwd: %s" % (cmd, cwd))

            if block:
                if capture_output:
                    # capture output
                    result = subprocess.run(cmd, cwd=cwd, shell=True, check=True, env=the_env, capture_output=True)
                    return result
                else:
                    # to output
                    result = subprocess.run(cmd, cwd=cwd, shell=True, check=True, env=the_env, capture_output=False)
                    return result
            else:
                result = subprocess.Popen(cmd, cwd=cwd, shell=True, env=the_env)
                return result

        except Exception as e:
            raise e


class CompletionMixin:
    def print_completion(self, descriptions, short_mode=False):
        if type(descriptions) == list:
            for one in descriptions:
                if type(one) == tuple or type(one) == list:
                    if not short_mode:
                        print(":".join(one))
                    else:
                        print(one[0])
                else:
                    print(one)



from setuptools import setup, find_packages
install_requires = [
    'requests==2.22.0',
    'dynaconf[all]==2.2.3',
    'pyfiglet==0.8.post1',
    'Click==7.0',
    'psutil',
    'kubernetes==11.0.0',
    'ruamel.yaml==0.16.12',
    'prettytable==0.7.2'
]


setup(
    name='oh-my-tools',
    version=0.1,
    description='oh my tools',
    license='MIT',
    author='Lu Ganlin',
    author_email='linewx1981@gmail.com',
    url='https://github.com/linewx/omt',
    packages=find_packages(),
    package_data={'omt.config': ['*.yaml'], 'omt.lib': ['**', '**/*', '**/**/*']},
    install_requires=install_requires,
    entry_points={
        'console_scripts': [
            'omt = omt.main:main',
        ],
    }
)



# -*- coding: utf-8 -*-

import os

import pkg_resources
from omt.config import settings

import omt
import logging
import copy
import textwrap

from omt.core.decorator import filecache


class Resource:
    # todo@rain: to differentiate between the following methods:
    # __methods private methods, don't inherit in child method
    # _methods, can be inherit in child method, but can't be invoked and public to end users
    # method, public method, can be invoke by the system
    def __init__(self, context={}, type='web'):
        self.context = context
        # set default resoure context
        self.context[self._get_resource_name()] = []
        self.logger = logging.getLogger('.'.join([self.__module__, self.__class__.__name__]))
        self.has_params = None
        self.type = type

    ################################################################
    #################### start method ##############################
    ################################################################
    # main method, system entrypoint

    def _exec(self):
        return self.__format(self.__execute())

    ################################################################
    #################### private method ############################
    ################################################################
    # private mehtod, only invoked in this class
    def __format(self, result):
        if self.type == 'cmd':
            print(result)
        elif self.type == 'web':
            return result
        else:
            raise Exception("unsupported type")

    def __execute(self):
        resource_name = self._get_resource_name()
        raw_command = self.context['all']
        index = self.context['index']
        params = raw_command[index + 1:]
        self.context[resource_name] = []  # init reosurce params

        while True:
            if not params:
                # has no params, stop parsing then start to run default action
                self._run()
                break

            else:
                # has params
                next_value = params[0]
                if next_value.startswith('-'):
                    # next value is resource params
                    self.context[resource_name].append(params[0])
                    self.context['index'] = self.context['index'] + 1  # increase index
                    index = self.context['index']
                    params = raw_command[index + 1:]
                elif next_value in self._get_submodules():
                    # chain to other resource

                    try:
                        # detect if it's a resource type
                        mod_path = (str(self.__class__.__module__)).split('.')[:-1]
                        mod_path.extend([next_value, next_value])
                        mod = __import__('.'.join(mod_path), fromlist=[next_value.capitalize()])
                        self._before_sub_resource()
                        context = copy.copy(self.context)
                        context['index'] = self.context['index'] + 1
                        context[resource_name] = self.context[resource_name]
                        if hasattr(mod, next_value.capitalize()):
                            clazz = getattr(mod, next_value.capitalize())
                            instance = clazz(context)
                            result = instance.__execute()
                            self._after_sub_resource()
                            return result
                        else:
                            self.logger.error("module %s deteched, but not exists in fact", next_value)
                            return
                    except ModuleNotFoundError as inst:
                        self.logger.info(inst)


                elif hasattr(self, next_value):
                    # next_value is resource action
                    action = getattr(self, next_value)
                    self.context['index'] += 1
                    self.context['action_params'] = raw_command[self.context['index'] + 1:]
                    return action()
                else:
                    #
                    # if self.has_params is None:
                    #     self.has_params = True
                    self.context['index'] = self.context['index'] + 1
                    index = self.context['index']
                    self.context[resource_name].append(next_value)
                    params = raw_command[index + 1:]

    ################################################################
    #################### system method ############################
    ################################################################
    # system methods, don't modify!!!
    def _get_resource_name(self):
        return self.__class__.__name__.lower()

    def _get_resource_values(self, resource_name=None):
        resource_name = resource_name if resource_name is not None else self._get_resource_name()
        return self.context[resource_name]

    def _get_one_resource_value(self, resource_name=None, index=0):
        resource_name = resource_name if resource_name is not None else self._get_resource_name()
        resource_values = self._get_resource_values(resource_name)
        if len(resource_values) <= index:
            return None
        else:
            return resource_values[index]

    def _have_resource_value(self):
        if self._get_resource_values():
            return True
        else:
            return False

    def _get_action_params(self):
        return self.context['action_params']

    def _get_params(self):
        raw_command = self.context['all']
        index = self.context['index']
        params = raw_command[index + 1:]
        return params

    def _get_public_methods(self):
        return list(filter(lambda x: callable(getattr(self, x)) and not x.startswith('_'), dir(self)))

    def _get_submodules(self):
        # for example, module name is 'omt.resources.jmx.jmx', submodules should be other folder within omt.resources.jmx folder
        module_path = self.__module__.split('.')
        module_path.pop()
        current_module = module_path.pop()
        all_resources = (pkg_resources.resource_listdir('.'.join(module_path), current_module))
        filterd_modules = [one for one in all_resources if
                           pkg_resources.resource_isdir('.'.join(module_path) + '.' + current_module,
                                                        one) is True and one not in ['__pycache__']]

        return filterd_modules

    def _print_completion(self, descriptions, short_mode=False):
        if type(descriptions) == list:
            for one in descriptions:
                if type(one) == tuple or type(one) == list:
                    if not short_mode:
                        print(":".join(one))
                    else:
                        print(one[0])
                else:
                    print(one)

    def _get_completion(self, descriptions, short_mode=False):
        result = []
        if type(descriptions) == list:
            for one in descriptions:
                if type(one) == tuple or type(one) == list:
                    if not short_mode:
                        result.append(":".join(one))
                    else:
                        result.append(one[0])
                else:
                    result.append(one)
        return result

    def _get_public_method_completion(self):
        public_methods = self._get_public_methods()
        return [(one, getattr(self, one).__doc__ if getattr(self, one).__doc__ is not None else one) for one in
                public_methods]

    def _get_sub_modules(self):
        return [(one_module, 'module ' + one_module) for one_module in self._get_submodules()]

    ################################################################
    #################### protected method ##########################
    ################################################################
    # default methods for children, can be overwrite in children

    # default method to run if  no action provided
    def _run(self):
        raise Exception('default action is not defined')

    def _before_sub_resource(self):
        pass

    def _after_sub_resource(self):
        pass

    def _before_method(self):
        pass

    def _after_method(self):
        pass

    def _description(self):
        """this is the description for resources"""
        return self._get_resource_name()

    # @staticmethod
    def _get_cache_file_name(self):
        main_path = [one for one in self.context['all'][1:] if not one.startswith('-')]
        cache_file = os.path.join(settings.OMT_COMPLETION_CACHE_DIR, *main_path)
        return cache_file
        # return '/tmp/file.txt'

    @filecache(duration=60 * 30, file=_get_cache_file_name)
    def _completion(self, short_mode=False):
        public_methods = self._get_public_method_completion()
        sub_modules = self._get_sub_modules()
        all_comp = []
        all_comp.extend(self._get_completion(public_methods, short_mode))
        all_comp.extend(self._get_completion(sub_modules, short_mode))
        return "\n".join(all_comp)

    def _help(self):
        raw_command = self.context['all']
        index = self.context['index']
        params = raw_command[index + 1:]
        if len(params) == 0:
            return textwrap.dedent(self.__doc__)
        try:
            first_param = params[0]
            # detect if it's a resource type
            mod_path = (str(self.__class__.__module__)).split('.')[:-1]
            mod_path.extend([first_param, first_param])
            mod = __import__('.'.join(mod_path), fromlist=[first_param.capitalize()])
            clazz = getattr(mod, first_param.capitalize())
            return clazz.textwrap.dedent(__doc__)
        except (ModuleNotFoundError, AttributeError) as inst:
            try:
                action = getattr(self, params[0])
                print(textwrap.dedent(action.__doc__))
            except Exception as inst:
                print(inst)

    ################################################################
    #################### public method #############################
    ################################################################
    # common public user methods, inherit protected methods if need change.

    def _clean_completin_cache(self):
        cache_file = self._get_cache_file_name()
        cache_folder = os.path.dirname(cache_file)

        duration_file = os.path.join(cache_folder, 'duration')

        if os.path.exists(cache_file):
            os.remove(cache_file)

        if os.path.exists(duration_file):
            os.remove(duration_file)

    def completion(self):
        if '--refresh' in self._get_action_params():
            self._clean_completin_cache()
        # list candidates for completions in zsh style (a:"description for a" b:"description for b")
        try:
            print(self._completion())
        except Exception as inst:
            # keep silent in completion mode
            return

    def help(self):
        self._help()

if __name__ == '__main__':
    Resource.__name__



import configparser
import logging


class ConfigReader:
    def __init__(self, cfg_file):
        self.cfg_file = cfg_file
        self.config_parser = configparser.RawConfigParser()
        self.config_parser.read(self.cfg_file)
        self.logger = logging.getLogger(__name__)


    def get_parser(self):
        return self.config_parser

    def get_sections(self):
        return self.config_parser.sections()

    def get_one_section(self, name):
        return self.config_parser[name]

    def get_with_section(self, section, key):
        print(self.config_parser[section][key])
        return self.config_parser[section][key]

    def get(self, key):
        if key is None:
            raise ValueError('key can\'t be empty')
        else:
            config_parser = configparser.ConfigParser()
            config_parser.read(self.cfg_file)
            try:
                if '.' in key:
                    key_list = key.split('.')
                    return config_parser[key_list[0]][key_list[1]]
                else:
                    return config_parser['default'][key]
            except Exception:
                self.logger.error('key %s not found' % key)

    def add(self, section, items):
        config_parser = configparser.RawConfigParser()
        config_parser.read(self.cfg_file)

        if section in config_parser.sections():
            raise ValueError('section %s already exist' % section)
        else:
            config_parser.add_section(section)
            for key, value in items.items():
                config_parser.set(section, key, value)

            with open(self.cfg_file, 'w') as configfile:
                config_parser.write(configfile)

    def update(self, section, items):
        config_parser = configparser.RawConfigParser()
        config_parser.read(self.cfg_file)

        if section not in config_parser.sections():
            raise ValueError('section %s not found' % section)
        else:
            config_parser.add_section(section)
            for key, value in items.items():
                config_parser.set(section, key, value)


            with open(self.cfg_file, 'w') as configfile:
                config_parser.write(configfile)